unit dumpActionsU;

interface

uses hostFilterU;

var g_pfMallocCallback: TMallocCallback;
    g_pfLogCallback: TLogCallback;
    g_wRunId: Cardinal;

function HTTHost_filter_proc_allow(_pstrucParams: PFilterProcParamStruc): Integer;
procedure HTTHost_filter_proc_connect(_pstrucParams: PFilterProcParamStruc);
function HTTHost_filter_proc_send(_pstrucParams: PFilterProcParamStruc): Integer;
function HTTHost_filter_proc_recv(_pstrucParams: PFilterProcParamStruc): Integer;
procedure HTTHost_filter_proc_disconnect(_pstrucParams: PFilterProcParamStruc);

implementation

uses Classes, Windows, SysUtils, dumpToolsU, Math;

type TConnectionInfo = class
                        constructor Create(const _strFileName: String);
                        destructor Destroy; override;
                       private
                        m_hFileHandle: THandle;
                       public
                        procedure Log(const _strLogLine: String);
                        procedure DumpData(const _strDirectionPrefix: String; const _ptrData: Pointer; const _intDataLength: Integer);
                       end;

var g_objConnections: TStringList;

//------------------------------------------------------------------------------

constructor TConnectionInfo.Create(const _strFileName: String);
begin
 inherited Create;
 m_hFileHandle := CreateFile(PChar(_strFileName), GENERIC_WRITE, FILE_SHARE_READ, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
end;

//------------------------------------------------------------------------------

procedure TConnectionInfo.Log(const _strLogLine: String);
var dwWr: Cardinal;
begin
 if m_hFileHandle <> INVALID_HANDLE_VALUE then WriteFile(m_hFileHandle, PChar(_strLogLine + #13#10)^, Length(_strLogLine) + 2, dwWr, nil);
end;

//------------------------------------------------------------------------------

procedure TConnectionInfo.DumpData(const _strDirectionPrefix: String; const _ptrData: Pointer; const _intDataLength: Integer);
const CHARS_PER_LINE = 16;
var intOffset, intIndex: Integer;
    strHexLine, strRawLine: String;
    bNextByte: Byte;

begin

 intOffset := 0;

 while intOffset < _intDataLength do
 begin

  strHexLine := '';
  strRawLine := '';

  for intIndex := 0 to Min(_intDataLength - intOffset, CHARS_PER_LINE) - 1 do
  begin
   bNextByte := Byte(Pointer(Cardinal(_ptrData) + Cardinal(intOffset) + Cardinal(intIndex))^);
   strHexLine := strHexLine + Format(' %.02x', [bNextByte]);
   if (Ord(bNextByte) < 32) or (Ord(bNextByte) > 127) then
    strRawLine := strRawLine + '.' else strRawLine := strRawLine + Chr(bNextByte); 
  end;

  Log(Format('%s |%-48s | %s', [_strDirectionPrefix, strHexLine, strRawLine]));

  Inc(IntOffset, CHARS_PER_LINE);
  
 end;
 
end;

//------------------------------------------------------------------------------

destructor TConnectionInfo.Destroy;
begin
 if m_hFileHandle <> INVALID_HANDLE_VALUE then CloseHandle(m_hFileHandle);
 inherited Destroy;
end;

//------------------------------------------------------------------------------
// connection attempt, use this handler for fast IP-based decisions.
// in:
// _pstrucParams^.dwConnectionId = connection seq id
// _pstrucParams^.intCallType = efcAllow
// _pstrucParams^.dwClientIP_n = proxy address, net format
// _pstrucParams^.dwTargetIP_n = target server address, net format
// _pstrucParams^.wTargetPort_n = target server port, net format
// returns:
// efrNoComment = proceed as usual
// efrDenyEverything = don't even attempt to connect, return failure to the client

function HTTHost_filter_proc_allow(_pstrucParams: PFilterProcParamStruc): Integer;
begin

 with _pstrucParams^ do
 begin

  // This is just an attempt, even if it's allowed, it can fail, and if it does
  // the plugin will not receive any more notifications, therefore we can't
  // allocate persistent TConnectionInfo here and simply log the attempt to the
  // HTTHost log. This also shows how to use that log.

  g_pfLogCallback(PChar(Format('dump.dll: connection attempt from proxy at %s to %s:%s',
                               [DWordToDottedIP(dwClientIP_n, afNet),
                                DWordToDottedIP(dwTargetIP_n, afNet),
                                WordToPortNumber(wTargetPort_n, afNet)])));

 end;

 Result := Integer(efrNoComment);

end;

//------------------------------------------------------------------------------
// connection is established, notification only. note, that once proc_connect
// is called, proc_disconnect WILL be called, no matter what happens with the
// connection, therefore use proc_disconnect for de-allocating any resources
// allocated in proc_connect.
// in:
// _pstrucParams^.dwConnectionId = connection seq id
// _pstrucParams^.intCallType = efcConnect
// _pstrucParams^.dwClientIP_n = proxy address, net format
// _pstrucParams^.dwTargetIP_n = target server address, net format
// _pstrucParams^.wTargetPort_n = target server port, net format

procedure HTTHost_filter_proc_connect(_pstrucParams: PFilterProcParamStruc);
var intIndex: Integer;
begin

 with _pstrucParams^ do
 begin

  // Connection has been established, allocate the new object, this creates a new log file.

  intIndex := g_objConnections.AddObject(IntToStr(dwConnectionId),
                                         TConnectionInfo.Create(Format('DUMP\htthost_%.04x_%.08x.log',
                                                                       [g_wRunId, dwConnectionId])));

  with TConnectionInfo(g_objConnections.Objects[intIndex]) do
  begin
   Log(Format('Connection from proxy at %s to %s:%s:',
              [DWordToDottedIP(dwClientIP_n, afNet),
               DWordToDottedIP(dwTargetIP_n, afNet),
               WordToPortNumber(wTargetPort_n, afNet)]));
  end;

 end;

end;

//------------------------------------------------------------------------------
// data is about to be sent to the target server, use this for data-based decisions,
// in-place data patching or data replacement
// in:
// _pstrucParams^.dwConnectionId = connection seq id
// _pstrucParams^.intCallType = efcSend
// _pstrucParams^.ptrData = pointer to the send data buffer
// _pstrucParams^.intDataLength = number of bytes that are to be sent
// _pstrucParams^.intAllocatedDataLength = number of bytes allocated in the buffer
// out (with retcode efrUseAnother):
// _pstrucParams^.ptrAnotherData
// _pstrucParams^.intAnotherDataLength
// _pstrucParams^.intAnotherAllocatedDataLength
// returns:
// efrNoComment = proceed to immediate sending. note, that you can modify the data
//                in the buffer as you please but the length may only be changed
//                with efrUseAnother.
// efrDenyEverything = send nothing, drop connection.
// efrDelaySend = delay sending until more data comes from the client, each time it does,
//                this call is repeated with the updated buffer content. return this
//                if there is not enough data in the buffer to make a decision (ex.
//                protocol-required CRLF has not been received from the client yet).
// efrAskMeLater = return this is the verification process is gonna take time
//                 (ex. requires database query for keywords). blocking server thread for
//                 more then several ms is undesirable, therefore you should start async.
//                 query (using async. ADO, your own separate thread or whatever),
//                 and return efrAskMeLater. The connection is blocked, no data is
//                 sent or received, HTTHost keeps calling again and again on each
//                 server thread pass (~ tens ms latency) until your async. query is
//                 complete, decision is made and any other code is returned.
// efrUseAnother = return this if you need to replace the whole data buffer.
//                 in this case allocate the buffer you need using g_pfMallocCallback,
//                 pour it with data, set _pstrucParams^.ptrAnotherData to the allocated
//                 buffer, _pstrucParams^.intAnotherDataLength to the length of the actual
//                 data, _pstrucParams^.intAnotherAllocatedDataLength to the size
//                 of the allocated buffer and return efrUseAnother. the original data
//                 is discarded and your data buffer is sent instead as if it was sent
//                 by the client.

function HTTHost_filter_proc_send(_pstrucParams: PFilterProcParamStruc): Integer;
var intIndex: Integer;
begin

 with _pstrucParams^ do
 if g_objConnections.Find(IntToStr(dwConnectionId), intIndex) then
 with TConnectionInfo(g_objConnections.Objects[intIndex]) do
 begin
  DumpData('>', ptrData, intDataLength);
 end;

 Result := Integer(efrNoComment);

end;

//------------------------------------------------------------------------------
// data has just been received from the target server, use this for data-based decisions,
// in-place data patching or data replacement
// in:
// _pstrucParams^.dwConnectionId = connection seq id
// _pstrucParams^.intCallType = efcRecv
// _pstrucParams^.ptrData = pointer to the received data buffer
// _pstrucParams^.intDataLength = number of bytes that are received and unconfirmed so far
// _pstrucParams^.intAllocatedDataLength = number of bytes allocated in the buffer
// out (with retcode efrUseAnother):
// _pstrucParams^.ptrAnotherData
// _pstrucParams^.intAnotherDataLength
// _pstrucParams^.intAnotherAllocatedDataLength
// returns:
// efrNoComment = data clearance is confirmed, data is copied to the client buffer,
//                from where the client will fetch it later with no further filter concern.
//                note, that you can modify the data in the buffer as you please, but the
//                length may only be changed with efrUseAnother.
// efrDenyEverything = confirm nothing, drop connection.
// efrDelaySend = delay confirmation until more data comes from the server, each time it does,
//                this call is repeated with the updated buffer content. return this
//                if there is not enough data in the buffer to make a decision (ex.
//                protocol-required CRLF has not been received from the client yet).
// efrAskMeLater = return this is the verification process is gonna take time
//                 (ex. requires database query for keywords). blocking server thread for
//                 more then several ms is undesirable, therefore you should start async.
//                 query (using async. ADO, your own separate thread or whatever),
//                 and return efrAskMeLater. The connection is blocked, no data is
//                 sent or received, HTTHost keeps calling again and again on each
//                 server thread pass (~ tens ms latency) until your async. query is
//                 complete, decision is made and any other code is returned.
// efrUseAnother = return this if you need to replace the whole data buffer.
//                 in this case allocate the buffer you need using g_pfMallocCallback,
//                 pour it with data, set _pstrucParams^.ptrAnotherData to the allocated
//                 buffer, _pstrucParams^.intAnotherDataLength to the length of the actual
//                 data, _pstrucParams^.intAnotherAllocatedDataLength to the size
//                 of the allocated buffer and return efrUseAnother. the original data
//                 is discarded and your data buffer is confirmed instead as if it was
//                 sent by the server.

function HTTHost_filter_proc_recv(_pstrucParams: PFilterProcParamStruc): Integer;
var intIndex: Integer;
begin

 with _pstrucParams^ do
 if g_objConnections.Find(IntToStr(dwConnectionId), intIndex) then
 with TConnectionInfo(g_objConnections.Objects[intIndex]) do
 begin
  DumpData('<', ptrData, intDataLength);
 end;

 Result := Integer(efrNoComment);

end;

//------------------------------------------------------------------------------
// connection is disconected, either upon proper protocol agreement between client
// and server, or due to network error, or even this very plugin returning eftDenyEverything.
// notification only. use proc_disconnect for de-allocating any resources
// allocated in proc_connect.
// in:
// _pstrucParams^.dwConnectionId = connection seq id
// _pstrucParams^.intCallType = efcDisconnect

procedure HTTHost_filter_proc_disconnect(_pstrucParams: PFilterProcParamStruc);
var intIndex: Integer;
begin

 with _pstrucParams^ do
 if g_objConnections.Find(IntToStr(dwConnectionId), intIndex) then
 with TConnectionInfo(g_objConnections.Objects[intIndex]) do
 begin
  Free;
  g_objConnections.Delete(intIndex);
 end;

end;

//------------------------------------------------------------------------------

initialization
 g_objConnections := TStringList.Create;
 g_objConnections.Sorted := True;
 g_objConnections.Duplicates := dupError;
finalization
 g_objConnections.Free;
end.
