unit dumpU;

interface

uses hostFilterU;

//------------------------------------------------------------------------------

procedure HTTHost_filter_init(const _pfMallocCallback, _pfLogCallback: Pointer); stdcall;
procedure HTTHost_filter_info(var _dwFilterFlags: Cardinal; const _pcBuffer: PChar; const _dwBufferSize: Cardinal); stdcall;
function HTTHost_filter_proc(_pstrucParams: PFilterProcParamStruc): Integer; stdcall;
procedure HTTHost_filter_done; stdcall;

//------------------------------------------------------------------------------

implementation

uses SysUtils, Windows, dumpActionsU;

const PLUGIN_DESCRIPTION = 'dumps all data being tunneled to separate files';
var g_sysobjLock: TRTLCriticalSection;

//------------------------------------------------------------------------------

procedure HTTHost_filter_init(const _pfMallocCallback, _pfLogCallback: Pointer);
begin
 g_pfMallocCallback := _pfMallocCallback;
 g_pfLogCallback := _pfLogCallback;
 Randomize;
 g_wRunId := Round(Now);
 InitializeCriticalSection(g_sysobjLock);
end;

//------------------------------------------------------------------------------

procedure HTTHost_filter_info(var _dwFilterFlags: Cardinal; const _pcBuffer: PChar; const _dwBufferSize: Cardinal);
begin
 _dwFilterFlags := FILTER_FLAG_CONNECT or FILTER_FLAG_RECV or FILTER_FLAG_SEND;
 StrLCopy(_pcBuffer, PLUGIN_DESCRIPTION, _dwBufferSize - 1);
end;

//------------------------------------------------------------------------------

function HTTHost_filter_proc(_pstrucParams: PFilterProcParamStruc): Integer;
begin

 EnterCriticalSection(g_sysobjLock);
 try

  case TExtensionFilterCall(_pstrucParams^.intCallType) of
  efcAllow:
  begin
   Result := HTTHost_filter_proc_allow(_pstrucParams);
  end;
  efcConnect:
  begin
   HTTHost_filter_proc_connect(_pstrucParams);
   Result := Integer(efrNoComment);
  end;
  efcSend:
  begin
   Result := HTTHost_filter_proc_send(_pstrucParams);
  end;
  efcRecv:
  begin
   Result := HTTHost_filter_proc_recv(_pstrucParams);
  end;
  efcDisconnect:
  begin
   HTTHost_filter_proc_disconnect(_pstrucParams);
   Result := Integer(efrNoComment);
  end;
  else
  begin
   g_pfLogCallback(PChar(Format('Unexpected command: %d', [_pstrucParams^.intCallType])));
   Result := Integer(efrDenyEverything);
  end;
  end;

 finally
  LeaveCriticalSection(g_sysobjLock);
 end;

end;

//------------------------------------------------------------------------------

procedure HTTHost_filter_done;
begin
 DeleteCriticalSection(g_sysobjLock);
end;

//------------------------------------------------------------------------------

end.
