unit httport_dllU;

interface

 uses Windows, Winsock, Classes, SysUtils;

//------------------------------------------------------------------------------

 // retcode constants

 const H_SUCCESS = 0;

 const H_ERROR_BASE = $10000;
 const H_ERROR_WINSOCK_FAILED = H_ERROR_BASE + 0;
 const H_ERROR_ALREADY_STARTED = H_ERROR_BASE + 1;
 const H_ERROR_NOT_STARTED = H_ERROR_BASE + 2;
 const H_ERROR_NO_FREE_SLOTS = H_ERROR_BASE + 3;
 const H_ERROR_UNKNOWN_OPTION = H_ERROR_BASE + 4;
 const H_ERROR_INVALID_PARAMETER = H_ERROR_BASE + 5;
 const H_ERROR_SOCKS_ALREADY_STARTED = H_ERROR_BASE + 6;
 const H_ERROR_SOCKS_NOT_STARTED = H_ERROR_BASE + 7;

 // h_setoption constants

 const H_OPTION_BASE = $20000;
 const H_OPTION_HOST_LIST = H_OPTION_BASE + 0;
 const H_OPTION_USER_INFO = H_OPTION_BASE + 1;
 const H_OPTION_PROXY = H_OPTION_BASE + 2;
 const H_OPTION_DATA_OVERRIDE = H_OPTION_BASE + 3;
 const H_OPTION_HEADER_FIELDS = H_OPTION_BASE + 4;

 // connection mode constants

 const CONN_USE_AUTO = 0;
 const CONN_USE_CONNECT = 1;
 const CONN_USE_REMOTE = 2;

//------------------------------------------------------------------------------

 // type definitions

 type TRetCode = Cardinal;

 type TMappingParameters = record                                              // this structure is passed to H_CreateMapping
                            intStrucSize: Integer;                             // [in] Must be set to sizeof(TMappingParameters)
                            wLocalPort: Word;                                  // [in] Local port to listen
                            pcRemoteHost: PChar;                               // [in] Pointer to ASCIIZ string with mapped host name or IP
                            wRemotePort: Word;                                 // [in] Mapped port
                            intAllowLocalOnly: Integer;                        // [in] If not zero, the listening port will be bound to 127.0.0.1
                            intMode: Integer;                                  // [in] Proxy bypass mode, one of CONN_USE_... constants
                            intMappingId: Integer;                             // [out] Returns the index for the newly created mapping
                           end;

 type PMappingParameters = ^TMappingParameters;

 type TOptions = array [0..$ffff] of PChar;                                    // structure for passing list of pointers to ASCIIZ strings to H_SetOption
      POptions = ^TOptions;

 type TMappingStats = record                                                   // this structure is passed to H_GetMappingStats
                       intStrucSize: Integer;                                  // [in] Must be set to sizeof(TMappingStats)
                       intConnections: Integer;                                // DEPRECATED
                       intKBytesTransferred: Integer;
                       intSockErrors: Integer;
                       intHostErrors: Integer;
                       intTimeoutErrors: Integer;
                       intProtocolErrors: Integer;
                       intMemoryErrors: Integer;
                       intConnectErrors: Integer;
                       intOtherErrors: Integer;
                       intFatalErrors: Integer;
                      end;

 type PMappingStats = ^TMappingStats;

 type TProxyStats = record                                                     // this structure is passed to H_GetHTTPStats
                     intStrucSize: Integer;                                    // [in] Must be set to sizeof(TProxyStats)
                     intProxyFaults: Integer;                                  // [out] DEPRECATED
                     intHostFaults: Integer;                                   // [out] DEPRECATED
                     intCurrentMode: Integer;                                  // [out] The 'worst' bypass mode currently in use for all mappings
                     intDataOverride: Integer;                                 // [out] If the proxy forbids passing custom request headers, bytes of data passed per URL
                    end;

 type PProxyStats = ^TProxyStats;

//------------------------------------------------------------------------------

 // function declarations

 type PCardinal = ^Cardinal;

 function H_Start: TRetCode; stdcall;
 function H_Stop: TRetCode; stdcall;

 function H_SetOption(const _intOption, _intCount: Integer; const _parrpcParams: POptions): TRetCode; stdcall;
 function H_CreateMapping(_pstrucParams: PMappingParameters): TRetCode; stdcall;
 function H_GetListeningPort(const _intIndex: Integer; _pdwResult: PCardinal): TRetCode; stdcall;
 function H_DestroyMapping(const _intIndex: Integer): TRetCode; stdcall;

 function H_GetMappingStats(const _intIndex: Integer; _pstrucStats: PMappingStats): TRetCode; stdcall; // DEPRECATED
 function H_GetProxyStats(const _pstrucStats: PProxyStats): TRetCode; stdcall;

 function H_CreateSocks(_pstrucParams: PMappingParameters): TRetCode; stdcall;
 function H_GetSocksListeningPort(_pdwResult: PCardinal): TRetCode; stdcall;
 function H_DestroySocks: TRetCode; stdcall;

 function H_ReadLog(var _pcLogLines: PChar): TRetCode; stdcall;
 function H_ReleaseMemory(_ptrRaw: Pointer): TRetCode; stdcall;

//------------------------------------------------------------------------------

implementation

 uses HTTPortGlobalsU, HTTPortMappingU, requestU, HTTPortHTTPU, HTTPortQueueU,
      HTTPortSocksU;

//------------------------------------------------------------------------------

 var g_boolInitialized: Boolean;
     g_arrobjMappings: array [0..MAX_PORTS-1] of TListenThread;
     g_objHTTPQueue: THTTPQueue;
     g_arrobjHTTPThreads: array [0..HTTP_THREADS-1] of THTTPThread;
     g_strProxyHost: String;
     g_wProxyPort: Word;
     g_strProxyHostPort: String;
     g_objSocksThread: TSocksThread;
     g_objKnownHosts: TStringList;
     g_strLogMessages: String;

//------------------------------------------------------------------------------

 var sysobjLock, sysobjLogLock: TRTLCriticalSection;

 procedure Lock;
 begin
  EnterCriticalSection(sysobjLock);
 end;

 procedure Unlock;
 begin
  LeaveCriticalSection(sysobjLock);
 end;

//------------------------------------------------------------------------------
// this callback function is called by all the worker threads for reporting
// problems instead of sending messages to the main window

 procedure Log(const _strMessage: String);
 begin
  EnterCriticalSection(sysobjLogLock);
  try
   if Length(g_strLogMessages) > 0 then g_strLogMessages := g_strLogMessages + #13#10;
   g_strLogMessages := g_strLogMessages + _strMessage;
  finally
   LeaveCriticalSection(sysobjLogLock);
  end;
 end;

//------------------------------------------------------------------------------
// returns the current contents of the error log, the log is then truncated

 function H_ReadLog(var _pcLogLines: PChar): TRetCode;
 begin

  EnterCriticalSection(sysobjLogLock);
  try

   GetMem(_pcLogLines, Length(g_strLogMessages) + 1);
   Move(PChar(g_strLogMessages)^, _pcLogLines^, Length(g_strLogMessages));
   _pcLogLines[Length(g_strLogMessages)] := #0;

   g_strLogMessages := '';
   Result := H_SUCCESS;

  finally
   LeaveCriticalSection(sysobjLogLock);
  end;

 end;

//------------------------------------------------------------------------------
// releases memory previously allocated in H_ReadLog

 function H_ReleaseMemory(_ptrRaw: Pointer): TRetCode;
 begin

  if not Assigned(_ptrRaw) then Result := H_ERROR_INVALID_PARAMETER
  else
  begin
   FreeMem(_ptrRaw);
   Result := H_SUCCESS;
  end;

 end;

//------------------------------------------------------------------------------
// initializes all the required structures, starts worker threads

 function H_Start: TRetCode;
 var intInd: Integer;
 begin

  Lock;
  try

   if g_boolInitialized then Result := H_ERROR_ALREADY_STARTED
   else
   begin

    if g_strProxyHostPort <> '' then
     HR_Init(PChar(g_strUserAgent), PChar(g_strProxyHostPort), PChar(g_strUsername), PChar(g_strPassword))
    else
     HR_Init(PChar(g_strUserAgent), nil, '', '');

    g_objHTTPQueue := THTTPQueue.Create;

    for intInd := 0 to HTTP_THREADS - 1 do
     g_arrobjHTTPThreads[intInd] := THTTPThread.Create('HTTP' + IntToStr(intInd + 1), Log, g_objHTTPQueue);

    g_strLogMessages := '';
    g_boolSwitchedToRemoteHostMode := False;

    g_boolInitialized := True;
    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// destroys socks and all mappings, stops all the worker threads and cleans up

 function H_Stop: TRetCode;
 var intInd: Integer;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   begin

    if Assigned(g_objSocksThread) then H_DestroySocks;

    for intInd := 0 to MAX_PORTS - 1 do
     if Assigned(g_arrobjMappings[intInd]) then H_DestroyMapping(intInd);

    for intInd := 1 to HTTP_DISCONNECT_DELAY do
     if g_objHTTPQueue.IsEmpty then Break else Sleep(1000);

    for intInd := 0 to HTTP_THREADS - 1 do g_arrobjHTTPThreads[intInd].Terminate;
    for intInd := 0 to HTTP_THREADS - 1 do
    begin
     g_arrobjHTTPThreads[intInd].WaitFor;
     g_arrobjHTTPThreads[intInd].Free;
    end;

    g_objHTTPQueue.Free;

    HR_Done;

    g_boolInitialized := False;
    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// sets various global options before the library is started with H_Start

 function H_SetOption(const _intOption, _intCount: Integer; const _parrpcParams: POptions): TRetCode;
 var intInd: Integer;
     intDataOverride: Integer;
 begin

  Lock;
  try

   if g_boolInitialized then Result := H_ERROR_ALREADY_STARTED                 // must be called before H_Start
   else
   if (_intCount <= 0) or (_intCount >= $10000) or (not Assigned(_parrpcParams)) then Result := H_ERROR_INVALID_PARAMETER
   else
   case _intOption of

   H_OPTION_HOST_LIST: //-------------------------------------------------------- set the list of knows htthosts
   begin                                                                       // takes (0..N-1) host:port
    g_objKnownHosts.Clear;
    for intInd := 0 to _intCount - 1 do g_objKnownHosts.Add(_parrpcParams^[intInd]);
    Result := H_SUCCESS;
   end;

   H_OPTION_USER_INFO: //-------------------------------------------------------- set user information to be shown to proxy
   if _intCount <> 3 then Result := H_ERROR_INVALID_PARAMETER                  // takes (0) user name, (1) password and (2) user-agent
   else
   begin
    g_strUsername := _parrpcParams^[0];
    g_strPassword := _parrpcParams^[1];
    g_strUserAgent := _parrpcParams^[2];
    g_boolProxyAuthentication := (Length(g_strUsername) > 0) or (Length(g_strPassword) > 0);
    Result := H_SUCCESS;
   end;

   H_OPTION_PROXY: //------------------------------------------------------------ set proxy host and port
   if _intCount <> 2 then Result := H_ERROR_INVALID_PARAMETER                  // takes (0) proxy host and (1) proxy port
   else
   if (StrLen(_parrpcParams^[0]) = 0) and (StrLen(_parrpcParams^[1]) = 0) then // no proxy - direct or firewall
   begin
    g_strProxyHost := '';
    g_wProxyPort := 0;
    g_strProxyHostPort := '';
    Result := H_SUCCESS;
   end
   else                                                                        // valid proxy
   try
    g_strProxyHost := _parrpcParams^[0];
    g_wProxyPort := StrToInt(_parrpcParams^[1]);
    g_strProxyHostPort := g_strProxyHost + ':' + IntToStr(g_wProxyPort);
    Result := H_SUCCESS;
   except
    Result := H_ERROR_INVALID_PARAMETER;
   end;

   H_OPTION_DATA_OVERRIDE: //---------------------------------------------------- set alternative mode for sending data (bytes in URL)
   if _intCount <> 1 then Result := H_ERROR_INVALID_PARAMETER                  // takes (0) number of bytes
   else
   try
    intDataOverride := StrToInt(_parrpcParams[0]);
    if (intDataOverride < 0) or (intDataOverride > 1024) then Result := H_ERROR_INVALID_PARAMETER
    else
    begin
     g_intDataOverride := intDataOverride;
     Result := H_SUCCESS;
    end;
   except
    Result := H_ERROR_INVALID_PARAMETER;
   end;

   H_OPTION_HEADER_FIELDS: //---------------------------------------------------- specify additional request header lines
   begin                                                                       // takes (0..N-1) Name: Value
    g_strAdditionalHeaders := _parrpcParams^[0];
    for intInd := 1 to _intCount - 1 do
     g_strAdditionalHeaders := g_strAdditionalHeaders + #13#10 + _parrpcParams^[intInd];  
    Result := H_SUCCESS;
   end;

   else Result := H_ERROR_UNKNOWN_OPTION; //------------------------------------- too bad
   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// creates a worker thread for mapping specified local port to the specified
// remote host's port

 function H_CreateMapping(_pstrucParams: PMappingParameters): TRetCode;
 var intInd: Integer;
     intMapIndex: Integer;

 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if not Assigned(_pstrucParams) then Result := H_ERROR_INVALID_PARAMETER
   else
   with _pstrucParams^ do
   if (intStrucSize <> SizeOf(TMappingParameters)) or
      ((intMode <> CONN_USE_AUTO) and (intMode <> CONN_USE_CONNECT) and (intMode <> CONN_USE_REMOTE)) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin

    g_boolSwitchedToRemoteHostMode := g_boolSwitchedToRemoteHostMode or (intMode = CONN_USE_REMOTE);

    intMapIndex := -1;

    for intInd := 0 to MAX_PORTS - 1 do
     if not Assigned(g_arrobjMappings[intInd]) then
     begin
      intMapIndex := intInd;
      Break;
     end;

    if intMapIndex = -1 then
    begin
     Result := H_ERROR_NO_FREE_SLOTS;
     Exit;
    end;

    g_arrobjMappings[intMapIndex] :=
       TListenThread.Create(Format('%s:%d', [pcRemoteHost, wRemotePort]),
                            g_strProxyHost, g_wProxyPort, wLocalPort, pcRemoteHost, wRemotePort,
                            Log, intMapIndex, g_objHTTPQueue, g_objKnownHosts, intAllowLocalOnly <> 0,
                            intMode, INVALID_SOCKET, '', '', 0);

    intMappingId := intMapIndex;
    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// whenever mapping was created with zero local port, this function should be used
// to fetch the actual value for the listening local port

 function H_GetListeningPort(const _intIndex: Integer; _pdwResult: PCardinal): TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if (_intIndex < 0) or (_intIndex >= MAX_PORTS) or (not Assigned(_pdwResult)) then Result := H_ERROR_INVALID_PARAMETER
   else
   if not Assigned(g_arrobjMappings[_intIndex]) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin
    _pdwResult^ := g_arrobjMappings[_intIndex].wListeningPort;                 // this can still return zero, if the thread had no chance to call bind() yet
    Result := H_SUCCESS;                                                       // but it normally takes several ms 
   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// this function is deprecated and returns all zeroes

 function H_GetMappingStats(const _intIndex: Integer; _pstrucStats: PMappingStats): TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   with _pstrucStats^ do
   if intStrucSize <> SizeOf(TMappingStats) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin

    intConnections := 0;
    intKBytesTransferred := 0;
    intSockErrors := 0;
    intHostErrors := 0;
    intTimeoutErrors := 0;
    intProtocolErrors := 0;
    intMemoryErrors := 0;
    intConnectErrors := 0;
    intOtherErrors := 0;
    intFatalErrors := 0;

    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// use this function just before you are stopping the library with H_Stop,
// it will return two highly relevant values - (1) actual tunneling mode to
// be used against currently chosen and proxy, and (2) DataOverride aka BytesInURL
// setting. Store the obtained values to some persistent storage, ex. registry
// and set them with when appropriate calls to H_SetOption next time you start
// the library    

 function H_GetProxyStats(const _pstrucStats: PProxyStats): TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if not Assigned(_pstrucStats) then Result := H_ERROR_INVALID_PARAMETER
   else
   with _pstrucStats^ do
   if intStrucSize <> SizeOf(TProxyStats) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin

    intProxyFaults := 0;
    intHostFaults := 0;

    intDataOverride := g_intDataOverride;                                      // bytes in URL is a global setting

    if g_boolSwitchedToRemoteHostMode then
     intCurrentMode := CONN_USE_REMOTE
    else
     intCurrentMode := CONN_USE_AUTO;                                          // return AUTO unless some mapping had to switch to remote

    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// stops and destroys the mapping worker thread

 function H_DestroyMapping(const _intIndex: Integer): TRetCode;
 begin

  Lock;
  try

   if (_intIndex < 0) or (_intIndex >= MAX_PORTS) then Result := H_ERROR_INVALID_PARAMETER
   else
   if not Assigned(g_arrobjMappings[_intIndex]) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin

    with g_arrobjMappings[_intIndex] do
    begin
     Terminate;
     WaitFor;
     Free;
    end;

    g_arrobjMappings[_intIndex] := nil;
    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// creates a single SOCKS4 server thread

 function H_CreateSocks(_pstrucParams: PMappingParameters): TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if not Assigned(_pstrucParams) then Result := H_ERROR_INVALID_PARAMETER
   else
   if Assigned(g_objSocksThread) then Result := H_ERROR_SOCKS_ALREADY_STARTED
   else
   with _pstrucParams^ do
   if (intStrucSize <> SizeOf(TMappingParameters)) or
      ((intMode <> CONN_USE_AUTO) and (intMode <> CONN_USE_CONNECT) and (intMode <> CONN_USE_REMOTE)) then Result := H_ERROR_INVALID_PARAMETER
   else
   begin
    g_boolSwitchedToRemoteHostMode := g_boolSwitchedToRemoteHostMode or (intMode = CONN_USE_REMOTE);
    g_objSocksThread := TSocksThread.Create(Log, wLocalPort, intAllowLocalOnly <> 0, intMode = CONN_USE_REMOTE,
                                            g_strProxyHost, g_wProxyPort, g_objHTTPQueue, intMode, g_objKnownHosts);
    Result := H_SUCCESS;
   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// whenever SOCKS server was created with zero local port, this function should be
// used to fetch the actual value for the listening local port

 function H_GetSocksListeningPort(_pdwResult: PCardinal): TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if not Assigned(_pdwResult) then Result := H_ERROR_INVALID_PARAMETER
   else
   if not Assigned(g_objSocksThread) then Result := H_ERROR_SOCKS_NOT_STARTED
   else
   begin
    _pdwResult^ := g_objSocksThread.wListeningPort;                            // this can still return zero, if the thread had no chance to call bind() yet
    Result := H_SUCCESS;                                                       // but it normally takes several ms
   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------
// stops and destroys SOCKS server thread

 function H_DestroySocks: TRetCode;
 begin

  Lock;
  try

   if not g_boolInitialized then Result := H_ERROR_NOT_STARTED
   else
   if not Assigned(g_objSocksThread) then Result := H_ERROR_SOCKS_NOT_STARTED
   else
   begin

    with g_objSocksThread do
    begin
     Terminate;
     WaitFor;
     Free;
    end;

    g_objSocksThread := nil;
    Result := H_SUCCESS;

   end;

  finally
   Unlock;
  end;

 end;

//------------------------------------------------------------------------------

var __strucWSAData: TWSAData;

initialization

 InitializeCriticalSection(sysobjLock);
 InitializeCriticalSection(sysobjLogLock);

 WSAStartup($101, __strucWSAData);

 g_objKnownHosts := TStringList.Create;
 g_objKnownHosts.Sorted := True;
 g_objKnownHosts.Duplicates := dupAccept;

finalization

 g_objKnownHosts.Free;

 WSACleanup;

 DeleteCriticalSection(sysobjLogLock);
 DeleteCriticalSection(sysobjLock);

end.

