#include "stdafx.h"
#include "HTTPort3.h"
#include <utility>
#include ".\httport3.h"
using std::make_pair;

#define _OLE2A(X) (reinterpret_cast<void*>(X) == 0 ? "" : OLE2A(X)) 

//-----------------------------------------------------------------------------

HRESULT CHTTPort3::HTTPORTCOMERROR(const HRetCode _hRetCode)
{
	if (_hRetCode == H_SUCCESS) return S_OK;
	if (_hRetCode == H_ERROR_WINSOCK_FAILED) return E_FAIL;
	if (_hRetCode == H_ERROR_ALREADY_STARTED) return E_UNEXPECTED;
	if (_hRetCode == H_ERROR_NOT_STARTED) return E_UNEXPECTED;
	if (_hRetCode == H_ERROR_NO_FREE_SLOTS) return E_OUTOFMEMORY;
	if (_hRetCode == H_ERROR_UNKNOWN_OPTION) return E_INVALIDARG;
	if (_hRetCode == H_ERROR_INVALID_PARAMETER) return E_INVALIDARG;
	if (_hRetCode == H_ERROR_SOCKS_ALREADY_STARTED) return E_UNEXPECTED;
	if (_hRetCode == H_ERROR_SOCKS_NOT_STARTED) return E_UNEXPECTED;
	return E_FAIL;
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::Start(BSTR _bsProxyHost, BSTR _bsProxyPort, 
							  BSTR _bsProxyUsername, BSTR _bsProxyPassword, 
							  BSTR _bsUserAgent, LONG _lTunnelingMode, LONG _lBytesInURL,
							  LONG _lAllowLocalOnly)
{
	USES_CONVERSION;
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		char *arrpcOptionArray[3];
		HRESULT hr;

		arrpcOptionArray[0] = _OLE2A(_bsProxyUsername);
		arrpcOptionArray[1] = _OLE2A(_bsProxyPassword);
		arrpcOptionArray[2] = _OLE2A(_bsUserAgent);
		hr = HTTPORTCOMERROR(h_setoption(H_OPTION_USER_INFO, 3, arrpcOptionArray));
		if (FAILED(hr)) return hr;

		arrpcOptionArray[0] = _OLE2A(_bsProxyHost);
		arrpcOptionArray[1] = _OLE2A(_bsProxyPort);
		hr = HTTPORTCOMERROR(h_setoption(H_OPTION_PROXY, 2, arrpcOptionArray));
		if (FAILED(hr)) return hr;

		char arrcBuf[16];
		wsprintf(arrcBuf, "%d", _lBytesInURL);
		arrpcOptionArray[0] = arrcBuf;
		hr = HTTPORTCOMERROR(h_setoption(H_OPTION_DATA_OVERRIDE, 1, arrpcOptionArray));
		if (FAILED(hr)) return hr;

		hr = HTTPORTCOMERROR(h_start());
		if (FAILED(hr)) return hr;
			
		m_lTunnelingMode = _lTunnelingMode;
		m_boolAllowLocalOnly = _lAllowLocalOnly != 0;

		m_boolStarted = true;
		return S_OK;

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::Stop(LONG* _plTunnelingMode, LONG* _plBytesInURL)
{
	try
	{
		
		if (!m_boolDLLOk) return E_FAIL;

		TProxyStats strucStats;
		strucStats.lStrucSize = sizeof(TProxyStats);
		
		HRESULT hr = HTTPORTCOMERROR(h_getproxystats(&strucStats));
		if (FAILED(hr)) return hr;

		*_plTunnelingMode = m_lTunnelingMode != CONN_USE_AUTO ? m_lTunnelingMode : strucStats.lCurrentMode;
		*_plBytesInURL = strucStats.lDataOverride;

		hr = HTTPORTCOMERROR(h_stop());
		if (FAILED(hr)) return hr;

		m_boolStarted = false;
		return S_OK;

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::CreateMapping(BSTR _bsTargetHost, LONG _lTargetPort, LONG _lLocalPort, LONG* _plLocalPort)
{
	USES_CONVERSION;
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		HRESULT hr;

		TMappingParameters strucMapping;
		strucMapping.lStrucSize = sizeof(TMappingParameters);
		strucMapping.lAllowLocalOnly = m_boolAllowLocalOnly ? 1 : 0;
		strucMapping.lMode = m_lTunnelingMode;
		strucMapping.pcRemoteHost = _OLE2A(_bsTargetHost);
		strucMapping.wRemotePort = static_cast<unsigned short>(_lTargetPort & 0xffff);
		strucMapping.wLocalPort = static_cast<unsigned short>(_lLocalPort & 0xffff);

		hr = HTTPORTCOMERROR(h_createmapping(&strucMapping));
		if (FAILED(hr)) return hr;

		*_plLocalPort = _lLocalPort;

		if (_lLocalPort == 0)
		{
			while (*_plLocalPort == 0)
			{
				Sleep(0);
				hr = HTTPORTCOMERROR(h_getlisteningport(strucMapping.lMappingId, _plLocalPort));
				if (FAILED(hr)) return hr;
			}
		}

		m_mapPortsToIndice.insert(make_pair(*_plLocalPort, strucMapping.lMappingId));

		return S_OK;

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::DestroyMapping(LONG _lLocalPort)
{
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		PortsToIndiceMap_t::iterator iterLookup = m_mapPortsToIndice.find(_lLocalPort);
		if (iterLookup == m_mapPortsToIndice.end()) return E_INVALIDARG;

		long lMappingIndex = iterLookup->second;
		m_mapPortsToIndice.erase(iterLookup);

		return HTTPORTCOMERROR(h_destroymapping(lMappingIndex));

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::StartSocks(LONG _lLocalPort, LONG* _plLocalPort)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		HRESULT hr;

		TMappingParameters strucMapping;
		strucMapping.lStrucSize = sizeof(TMappingParameters);
		strucMapping.lAllowLocalOnly = m_boolAllowLocalOnly ? 1 : 0;
		strucMapping.lMode = m_lTunnelingMode;
		strucMapping.wLocalPort = static_cast<unsigned short>(_lLocalPort & 0xffff);

		hr = HTTPORTCOMERROR(h_createsocks(&strucMapping));
		if (FAILED(hr)) return hr;

		*_plLocalPort = _lLocalPort;

		if (_lLocalPort == 0)
		{
			while (*_plLocalPort == 0)
			{
				Sleep(0);
				hr = HTTPORTCOMERROR(h_getsockslisteningport(_plLocalPort));
				if (FAILED(hr)) return hr;
			}
		}

		return S_OK;

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::StopSocks(void)
{
	try
	{
		if (!m_boolDLLOk) return E_FAIL;
		return HTTPORTCOMERROR(h_destroysocks());
	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::ReadLog(BSTR* _pbsLogLines)
{
	USES_CONVERSION;
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		char* pcLogLines;

		HRESULT hr = HTTPORTCOMERROR(h_readlog(&pcLogLines));
		if (FAILED(hr)) return hr;

		*_pbsLogLines = SysAllocString(A2OLE(pcLogLines));
		return S_OK;

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::SetHostList(BSTR _bsHostColonPortCrlfSeparated)
{
	USES_CONVERSION;
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		char* pcHostColonPortCrlfSeparated = _OLE2A(_bsHostColonPortCrlfSeparated);
		char* pcScan = pcHostColonPortCrlfSeparated;

		const int iMaxHosts = 256;
		char *arrpcOptionArray[iMaxHosts];
		int i;

		for (i = 0; i < iMaxHosts; ++i)
		{
			arrpcOptionArray[i] = pcScan;
			char* pcEScan = strstr(pcScan, "\r\n");
			if (pcEScan == 0) break;
			*pcEScan = '\0';
			pcScan = pcEScan + 2;
		}

		return HTTPORTCOMERROR(h_setoption(H_OPTION_HOST_LIST, i + 1, arrpcOptionArray));

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

STDMETHODIMP CHTTPort3::SetHeaderFields(BSTR _bsNameColonSpaceValueCrlfSeparated)
{
	USES_CONVERSION;
	try
	{

		if (!m_boolDLLOk) return E_FAIL;

		char* pcNameColonSpaceValueCrlfSeparated = _OLE2A(_bsNameColonSpaceValueCrlfSeparated);
		char* pcScan = pcNameColonSpaceValueCrlfSeparated;

		const int iMaxHosts = 256;
		char *arrpcOptionArray[iMaxHosts];
		int i;

		for (i = 0; i < iMaxHosts; ++i)
		{
			arrpcOptionArray[i] = pcScan;
			char* pcEScan = strstr(pcScan, "\r\n");
			if (pcEScan == 0) break;
			*pcEScan = '\0';
			pcScan = pcEScan + 2;
		}

		return HTTPORTCOMERROR(h_setoption(H_OPTION_HEADER_FIELDS, i + 1, arrpcOptionArray));

	}
	catch (...)
	{
		return E_FAIL;
	}
}

//-----------------------------------------------------------------------------

