// HTTPort version 3
// TCP/IP over HTTP tunneling client/server project
// (c) Dmitry Dvoinikov

// version 3.SNFM. Dmitry Dvoinikov
// This unit has been copied from HTTPortWinInetU.pas and modified so that
// all the parameters are accessible to the exported function calls.

unit requestU;

interface

 const HTTP_RETCODE_WININET_FAULT = 999;

 procedure HR_Init(const _pcUserAgent, _pcProxyHostPort, _pcUsername, _pcPassword: PChar); stdcall; export;

 function HR_SyncRequest(const _pcURL, _pcAdditionalHeaders: PChar; const _pcReferer: PChar;
                         var _wRetCode: Word; var _pcRetMessage: PChar;
                         var _intDataLength: Integer): Pointer; stdcall; export;

 procedure HR_ReleaseMemory(_ptrRaw: Pointer); stdcall; export;
 
 procedure HR_Done; stdcall; export;


implementation

 uses Classes, SysUtils, WinInet, Windows, requestwininetU, requestutilitiesU;

//------------------------------------------------------------------------------
// global variables

 const RETURN_SIZE_LIMIT = 1048576;

 var g_objWinInetConnections: TStringList;        // connection cache
 var g_sysobjWinInetLock: TRTLCriticalSection;    // connection cache lock
 var g_hInet: HInternet;                          // application internet handle

 var g_strUsername: String;
     g_strPassword: String;

 const g_arrAcceptedTypes: array [0..3] of PChar = ('text/*', 'image/*', 'application/*', nil); // what types of results to accept

//------------------------------------------------------------------------------
// takes 'User_Agent/V.VV', 'proxy.host.or.ip:port', 'username' and 'pa55w0rd'
// initializes library, returns nothing

 procedure HR_Init(const _pcUserAgent, _pcProxyHostPort, _pcUsername, _pcPassword: PChar);
 begin

  InitializeCriticalSection(g_sysobjWinInetLock);

  if _pcProxyHostPort = nil then
   g_hInet := InternetOpen(_pcUserAgent, INTERNET_OPEN_TYPE_DIRECT, nil, nil, 0)
  else
   g_hInet := InternetOpen(_pcUserAgent, INTERNET_OPEN_TYPE_PROXY, _pcProxyHostPort, nil, 0);


  g_objWinInetConnections := TStringList.Create;
  g_objWinInetConnections.Sorted := True;
  g_objWinInetConnections.Duplicates := dupError;

  g_strUsername := _pcUsername;
  g_strPassword := _pcPassword;

 end;

//------------------------------------------------------------------------------
// cleans library up, returns nothing

 procedure HR_Done;
 var intInd: Integer;
 begin

  for intInd := 0 to g_objWinInetConnections.Count - 1 do
   TWinInetConnection(g_objWinInetConnections.Objects[intInd]).Free;
  g_objWinInetConnections.Free;

  if g_hInet <> nil then
  begin
   InternetCloseHandle(g_hInet);
   g_hInet := nil;
  end;

  DeleteCriticalSection(g_sysobjWinInetLock);

 end;

//------------------------------------------------------------------------------
// in
// _pcURL -> 'http://host.name.or.ip[:port]/path?query'
// _pcAdditionalHeaders -> 'First-Header-Line: Value'CRLF...CRLF'Last-Header-Line: Value',
// _pcReferer -> 'http://host.name.or.ip:port/path?query'
//
// runs a GET request to the specified host through a proxy
//
// out
// _wRetCode = return HTTP code or HTTP_RETCODE_WININET_FAULT
// _pcRetMessage = nil if _wRetCode = 200, allocated message otherwise, raw binary data
// _intDataLength = result length, raw binary bytes 
// returns
// nil if _wRetCode <> 200, allocated result data otherwise
//
// this function must be thread-safe

 function HR_SyncRequest(const _pcURL, _pcAdditionalHeaders: PChar; const _pcReferer: PChar;
                         var _wRetCode: Word; var _pcRetMessage: PChar;
                         var _intDataLength: Integer): Pointer; stdcall; export;

 var strHostPort, strL, strR, strQuery: String;
     hConn, hReq: HInternet;
     intIndex: Integer;
     objNewConn: TWinInetConnection;
     dwContext: Cardinal;
     dwRetCode, dwContentLength, dwBufSize: Cardinal;
     dwToReadTotal, dwAvail, dwRead: Cardinal;

  procedure SetRetMessage(var _pcRetMessage: PChar; const _strRetMessage: String);
  begin
   GetMem(_pcRetMessage, Length(_strRetMessage) + 1);
   Move(PChar(_strRetMessage)^, _pcRetMessage^, Length(_strRetMessage));
   _pcRetMessage[Length(_strRetMessage)] := #0;
  end;

 begin

  Result := nil;
  _intDataLength := 0;
  _wRetCode := HTTP_RETCODE_WININET_FAULT;

  if g_hInet = nil then // not initialized
  begin
   SetRetMessage(_pcRetMessage, 'WinInet not initialized');
   Exit;
  end;

  SplitBy(_pcURL, 'http://', strL, strR);
  SplitBy(strR, '/', strHostPort, strQuery);

  // attempt to find a connection to this host in the connection cache

  EnterCriticalSection(g_sysobjWinInetLock);
  try

   if g_objWinInetConnections.Find(strHostPort, intIndex) then
    hConn := TWinInetConnection(g_objWinInetConnections.Objects[intIndex]).hConnectionHandle
   else
   begin // not found, create new connection and put it to cache
    objNewConn := TWinInetConnection.Create(g_hInet, strHostPort, g_strUsername, g_strPassword);
    g_objWinInetConnections.AddObject(strHostPort, objNewConn);
    hConn := objNewConn.hConnectionHandle;
   end;

  finally
   LeaveCriticalSection(g_sysobjWinInetLock);
  end;

  if hConn = nil then // too bad, failed to initialize connection
  begin
   SetRetMessage(_pcRetMessage, 'WinInet no connection');
   Exit;
  end;

  // create HTTP request

  dwContext := 0; 
  hReq := HTTPOpenRequest(hConn, nil, PChar('/' + strQuery), 'HTTP/1.1', _pcReferer,
                          @g_arrAcceptedTypes,
                          INTERNET_FLAG_KEEP_CONNECTION or INTERNET_FLAG_NO_CACHE_WRITE or
                          INTERNET_FLAG_PRAGMA_NOCACHE, dwContext);
  if hReq = nil then
  begin
   SetRetMessage(_pcRetMessage, Format('WinInet HTTPOpenRequest failed (%d)', [GetLastError]));
   Exit;
  end;

  try

   // send request to the proxy

   if not HTTPSendRequest(hReq, _pcAdditionalHeaders, Cardinal(-1), nil, 0) then
   begin
    SetRetMessage(_pcRetMessage, Format('WinInet HTTPSendRequest failed (%d)', [GetLastError]));
    Exit;
   end;

   // fetch HTTP retcode

   dwBufSize := SizeOf(dwRetCode);

   if not HTTPQueryInfo(hReq, HTTP_QUERY_STATUS_CODE or HTTP_QUERY_FLAG_NUMBER,
                        @dwRetCode, dwBufSize, Cardinal(Pointer(nil)^)) then
   begin
    SetRetMessage(_pcRetMessage, Format('WinInet HTTPQueryInfo(HTTP_QUERY_STATUS_CODE) failed (%d)', [GetLastError]));
    Exit;
   end;
   
   _wRetCode := dwRetCode;

   if _wRetCode <> 200 then
   begin
    SetRetMessage(_pcRetMessage, Format('HTTP request returned retcode %d (%s)', [_wRetCode, GetHTTPErrorDescription(_wRetCode)]));
    Exit;
   end;

   // fetch HTTP returned Content-Length

   dwBufSize := SizeOf(dwContentLength);

   if not HTTPQueryInfo(hReq, HTTP_QUERY_CONTENT_LENGTH or HTTP_QUERY_FLAG_NUMBER,
                        @dwContentLength, dwBufSize, Cardinal(Pointer(nil)^)) then
   begin
    _wRetCode := HTTP_RETCODE_WININET_FAULT;
    SetRetMessage(_pcRetMessage, Format('WinInet HTTPQueryInfo(HTTP_QUERY_CONTENT_LENGTH) failed (%d)', [GetLastError]));
    Exit;
   end;

   if (dwContentLength > RETURN_SIZE_LIMIT) or (dwContentLength = 0) then
   begin
    _wRetCode := HTTP_RETCODE_WININET_FAULT;
    SetRetMessage(_pcRetMessage, Format('Request returned either too much or no data, %d bytes', [dwContentLength]));
    Exit;
   end;

   // allocate result memory

   _intDataLength := Integer(dwContentLength);
   GetMem(Result, _intDataLength);

   // read all the result data

   dwToReadTotal := dwContentLength;

   while dwToReadTotal > 0 do
   try

    if not InternetQueryDataAvailable(hReq, dwAvail, 0, 0) then // fetch data available for reading, this call BLOCKS
     raise Exception.CreateFmt('WinInet InternetQueryDataAvailable failed (%d)', [GetLastError]);

    if (dwAvail > dwToReadTotal) or (dwAvail = 0) then // someone is cheating on us
     raise Exception.CreateFmt('Content-Length mismatch, expected (at most) %d bytes, returned %d bytes', [dwToReadTotal, dwAvail]);

    if not InternetReadFile(hReq, Pointer(Cardinal(Result) + dwContentLength - dwToReadTotal), dwAvail, dwRead) then
     raise Exception.CreateFmt('WinInet InternetReadFile failed (%d)', [GetLastError]);

    if dwRead <> dwAvail then // again, someone is cheating
     raise Exception.CreateFmt('WinInet InternetReadFile read %d bytes of %d available', [dwRead, dwAvail]);

    Dec(dwToReadTotal, dwRead);

   except
    FreeMem(Result);
    Result := nil;
    _wRetCode := HTTP_RETCODE_WININET_FAULT;
    _intDataLength := 0;
    SetRetMessage(_pcRetMessage, Exception(ExceptObject).Message);
    Exit;
   end;

  finally
   InternetCloseHandle(hReq);
  end;

 end;

//------------------------------------------------------------------------------

 procedure HR_ReleaseMemory(_ptrRaw: Pointer);
 begin
  if _ptrRaw <> nil then FreeMem(_ptrRaw);
 end;

//------------------------------------------------------------------------------

end.
