unit transferhtmlU;

interface

//------------------------------------------------------------------------------

 // HTTHost specific
 // takes original client URL by which current request is being made, and seven
 // null-terminated lines in BASE64 alphabet containing different response parts,
 // allocates, composes and returns partial HTTP response, in a raw memory block

 function HX_ComposeHTTHostHTTPResponse(const _pcRequestURL: PChar;
                                        const _pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7: PChar;
                                        var _intResponseLength: Integer): Pointer; stdcall; export;

 // HTTPort specific
 // takes original client URL by which current request is being made, and seven
 // null-terminated lines in BASE64 alphabet containing different response parts,
 // allocates, composes and returns fake response body, in a raw memory block

 function HX_ComposeHTTHostResponseBody(const _pcRequestURL: PChar;
                                        const _pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7: PChar;
                                        var _intResponseLength: Integer): Pointer; stdcall; export;

 // HTTPort specific
 // takes the URL response by which has just been returned, pointer to and size of
 // the returned response content
 // returns 1 if the response is valid, 0 otherwise

 function HX_IsValidHTTHostResponse(const _pcRequestURL: PChar;
                                    const _ptrResponseData: Pointer;
                                    const _intResponseDataLength: Integer): Integer; stdcall; export;

 // HTTPort specific
 // takes the URL response by which has just been returned, pointer to and size of
 // the returned response content and field index in range 1-7.
 // fetches, allocates and returns the value of the line with the specified index

 function HX_ExtractHTTHostResponse(const _pcRequestURL: PChar;
                                    const _ptrResponseData: Pointer;
                                    const _intResponseDataLength: Integer;
                                    const _intIndex: Integer): PChar; stdcall; export;

//------------------------------------------------------------------------------

implementation

 uses SysUtils, transferU;

 const CRLF = #13#10;

 var strTag1, strTag2, strTag3, strTag4, strTag5, strTag6, strTag7: String;

//------------------------------------------------------------------------------

 function ComposeHTTHostResponseBodyInternal(const _pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7: PChar): String;
 begin
  Result := Format('<html><head><%s><%s><%s><%s><%s><%s><%s></head><body>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '<%s>%s</%s>' +
                   '</body></html>',
                   [strTag1, strTag2, strTag3, strTag4, strTag5, strTag6, strTag7,
                   strTag1, _pcLine1, strTag1,
                   strTag2, _pcLine2, strTag2,
                   strTag3, _pcLine3, strTag3,
                   strTag4, _pcLine4, strTag4,
                   strTag5, _pcLine5, strTag5,
                   strTag6, _pcLine6, strTag6,
                   strTag7, _pcLine7, strTag7]);
 end;

//------------------------------------------------------------------------------

 function HX_ComposeHTTHostHTTPResponse(const _pcRequestURL: PChar; const _pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7: PChar;
                                        var _intResponseLength: Integer): Pointer;
 var strResult: String;
 begin

  Result := nil;
  
  try

   if (_pcRequestURL = nil) or (_pcLine1 = nil) or (_pcLine2 = nil) or (_pcLine3 = nil) or
      (_pcLine4 = nil) or (_pcLine5 = nil) or (_pcLine6 = nil) or (_pcLine7 = nil) then Exit;

   // Note to implementers: when HX_ComposeHTTHostHTTPResponse is called, all _pcLineX point
   // to some null-terminated ASCII lines. You may treat them as identical but here's
   // a hint: _pcLine3 would normally contain data portion and
   // therefore be the biggest of all, _pcLine5, _pcLine6 and _pcLine7 are obsoleted and
   // would normally be empty (but don't count on it) and the rest would point to lines
   // few bytes long.

   // Note to implementers: you can use URL to decide what to return, ex. create
   // a fake JPG for a URL that ends with '.jpg' etc. Then make sure the other
   // two functions analyze the URL too.

   // Note to implementers: what you are returning will be prepended with
   // -----
   // HTTP/1.1 200 OK
   // Connection: keep-alive OR close (HTTHost decides)
   // Cache-Control: no-cache
   // Pragma: no-cache
   // -----
   // and sent to the client. You may add whatever fields you need and/or return
   // textual or binary data, ex. JPG or GIF, in which case mind setting Content-Type
   // properly. Also note that HTTPort will only accept text/*, image/* and application/*
   // or see request.dll 

   strResult := ComposeHTTHostResponseBodyInternal(_pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7);

   strResult := Format('Content-Type: text/html' + CRLF +
                       'Content-Length: %d' + CRLF +
                       CRLF +
                       '%s',
                       [Length(strResult), strResult]);

   _intResponseLength := Length(strResult);
   Result := HX_AllocateMemory(_intResponseLength);
   Move(PChar(strResult)^, Result^, _intResponseLength);

  except
   Result := nil;
  end;

 end;

//------------------------------------------------------------------------------

 function HX_ComposeHTTHostResponseBody(const _pcRequestURL: PChar; const _pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7: PChar;
                                        var _intResponseLength: Integer): Pointer;
 var strResult: String;
 begin

  Result := nil;

  try

   if (_pcRequestURL = nil) or (_pcLine1 = nil) or (_pcLine2 = nil) or (_pcLine3 = nil) or
      (_pcLine4 = nil) or (_pcLine5 = nil) or (_pcLine6 = nil) or (_pcLine7 = nil) then Exit;

   // Note to implementers: HX_ComposeHTTHostResponseBody is only called by HTTPort
   // (that's right, the CLIENT !) to obtain a fake erroneous response, so you
   // may come up with something really simple, the result of this call will not
   // travel through proxy. For exactly the same reason, this function must return
   // just the body of the page (or binary data), no HTTP response header fields here. 

   strResult := ComposeHTTHostResponseBodyInternal(_pcLine1, _pcLine2, _pcLine3, _pcLine4, _pcLine5, _pcLine6, _pcLine7);

   _intResponseLength := Length(strResult);
   Result := HX_AllocateMemory(_intResponseLength);
   Move(PChar(strResult)^, Result^, _intResponseLength);

  except
   Result := nil;
  end;

 end;

//------------------------------------------------------------------------------

 function ExtractTagContents(var _pcSource: PChar; const _strLeft, _strRight: String): String;
 var pcScan, pcEScan: PChar;
 begin

  Result := '';
  if _pcSource = nil then raise Exception.Create('ExtractTagContents(): _pcSource = nil');

  pcScan := StrPos(_pcSource, PChar(_strLeft));
  if pcScan = nil then raise Exception.CreateFmt('ExtractTagContents(): no ''%s'' found', [_strLeft]);
  Inc(pcScan, Length(_strLeft));

  pcEScan := StrPos(pcScan, PChar(_strRight));
  if pcEScan = nil then raise Exception.CreateFmt('ExtractTagContents(): no ''%s'' found', [_strRight]);

  Result := StringOfChar(#0, pcEScan - pcScan);
  Move(pcScan^, PChar(Result)^, pcEScan - pcScan);

  _pcSource := pcEScan + Length(_strRight);

 end;

//------------------------------------------------------------------------------

 function HX_IsValidHTTHostResponse(const _pcRequestURL: PChar;
                                    const _ptrResponseData: Pointer;
                                    const _intResponseDataLength: Integer): Integer;
 var strResponse: String;
     strHead, strBody: String;
     strTag1, strTag2, strTag3, strTag4, strTag5, strTag6, strTag7: String;
     pcParse: PChar;
 begin

  try

   Result := 0;
   if (_pcRequestURL = nil) or (_ptrResponseData = nil) or (_intResponseDataLength <= 0) then Exit;

   strResponse := StringOfChar(#0, _intResponseDataLength);
   Move(_ptrResponseData^, PChar(strResponse)^, _intResponseDataLength);

   pcParse := PChar(strResponse);
   strHead := ExtractTagContents(pcParse, '<head>', '</head>');
   strBody := ExtractTagContents(pcParse, '<body>', '</body>');

   pcParse := PChar(strHead);
   strTag1 := ExtractTagContents(pcParse, '<', '>');
   strTag2 := ExtractTagContents(pcParse, '<', '>');
   strTag3 := ExtractTagContents(pcParse, '<', '>');
   strTag4 := ExtractTagContents(pcParse, '<', '>');
   strTag5 := ExtractTagContents(pcParse, '<', '>');
   strTag6 := ExtractTagContents(pcParse, '<', '>');
   strTag7 := ExtractTagContents(pcParse, '<', '>');

   pcParse := PChar(strBody);
   ExtractTagContents(pcParse, '<' + strTag1 + '>', '</' + strTag1 + '>');
   ExtractTagContents(pcParse, '<' + strTag2 + '>', '</' + strTag2 + '>');
   ExtractTagContents(pcParse, '<' + strTag3 + '>', '</' + strTag3 + '>');
   ExtractTagContents(pcParse, '<' + strTag4 + '>', '</' + strTag4 + '>');
   ExtractTagContents(pcParse, '<' + strTag5 + '>', '</' + strTag5 + '>');
   ExtractTagContents(pcParse, '<' + strTag6 + '>', '</' + strTag6 + '>');
   ExtractTagContents(pcParse, '<' + strTag7 + '>', '</' + strTag7 + '>');

   // if any of the above extract's fails, exception is thrown
   // otherwise we are still here therefore the structure is valid

   Result := 1;

  except
   Result := 0;
  end;

 end;

//------------------------------------------------------------------------------

 function HX_ExtractHTTHostResponse(const _pcRequestURL: PChar;
                                    const _ptrResponseData: Pointer;
                                    const _intResponseDataLength: Integer;
                                    const _intIndex: Integer): PChar;
 var strResponse, strHead, strBody, strTag, strValue: String;
     pcParse: PChar;
     intInd: Integer;
 begin

  try

   Result := nil;
   if (_pcRequestURL = nil) or (_ptrResponseData = nil) or (_intResponseDataLength <= 0) or
      (_intIndex < 1) or (_intIndex > 7) then Exit;

   strResponse := StringOfChar(#0, _intResponseDataLength);
   Move(_ptrResponseData^, PChar(strResponse)^, _intResponseDataLength);
   
   pcParse := PChar(strResponse);
   strHead := ExtractTagContents(pcParse, '<head>', '</head>');
   strBody := ExtractTagContents(pcParse, '<body>', '</body>');

   pcParse := PChar(strHead);
   for intInd := 1 to _intIndex do
   begin
    strTag := ExtractTagContents(pcParse, '<', '>');
   end;

   pcParse := PChar(strBody);
   strValue := ExtractTagContents(pcParse, '<' + strTag + '>', '</' + strTag + '>');

   Result := HX_AllocateMemory(Length(strValue) + 1);
   Move(PChar(strValue)^, Result^, Length(strValue));
   Result[Length(strValue)] := #0;

  except
   Result := nil;
  end;

 end;

//------------------------------------------------------------------------------

 function CreateRandomString(const _intLength: Integer): String;
 var intInd: Integer;
 begin
  Result := '';
  for intInd := 1 to _intLength do Result := Result + Chr(Ord('a') + Random(26));
 end;

//------------------------------------------------------------------------------

 procedure CreateRandomTags;
 begin
  strTag1 := CreateRandomString(4 + Random(4));
  repeat
   strTag2 := CreateRandomString(4 + Random(4));
  until (strTag2 <> strTag1);
  repeat
   strTag3 := CreateRandomString(4 + Random(4));
  until (strTag3 <> strTag2) and (strTag3 <> strTag1);
  repeat
   strTag4 := CreateRandomString(4 + Random(4));
  until (strTag4 <> strTag3) and (strTag4 <> strTag2) and (strTag4 <> strTag1);
  repeat
   strTag5 := CreateRandomString(4 + Random(4));
  until (strTag5 <> strTag4) and (strTag5 <> strTag3) and (strTag5 <> strTag2) and (strTag5 <> strTag1);
  repeat
   strTag6 := CreateRandomString(4 + Random(4));
  until (strTag6 <> strTag5) and (strTag6 <> strTag4) and (strTag6 <> strTag3) and (strTag6 <> strTag2) and (strTag6 <> strTag1);
  repeat
   strTag7 := CreateRandomString(4 + Random(4));
  until (strTag7 <> strTag6) and (strTag7 <> strTag5) and (strTag7 <> strTag4) and (strTag7 <> strTag3) and (strTag7 <> strTag2) and (strTag7 <> strTag1);
 end;

//------------------------------------------------------------------------------

initialization
 Randomize;
 CreateRandomTags;
end.
