unit transferurlU;

interface

//------------------------------------------------------------------------------

 // HTTPort specific
 // takes pointer to the asciiz data to hide in the URI (path + '?' + query)
 // allocates and returns URI with hidden data, also allocates and returns
 // a line to be passed in the Referer request header field

 function HX_EncodeURL(const _pcData: PChar; var _pcReferer: PChar): PChar; stdcall; export;

 // HTTHost specific
 // takes pointer to encoded URL, decodes, allocates and returns original asciiz data
 // referer is ignored

 function HX_DecodeURL(const _pcURL: PChar): PChar; stdcall; export;

//------------------------------------------------------------------------------

implementation

 uses SysUtils, Windows, transferU;

//------------------------------------------------------------------------------
// these are the first 64 prime numbers used for URL scrambling

 const URLPrimes: array [0..63] of Integer =
 (
 2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
 109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,223,227,
 229,233,239,241,251,257,263,269,271,277,281,283,293,307,311
 );

//------------------------------------------------------------------------------
// the following forward and reverse encoding maps are almost identical to BASE64,
// except & is used instead of + to make it more suitable for URLs

 const URL64MAP: array [0..63] of Char =
 (
 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
 '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '&', '/'
 );

 const _TR = $80;
 const _EN = $80;
 const _SK = $81;
 const _EQ = $82;

 const URL64MAP_REV: array [0..255] of Byte =
 (
 _EN, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,    // 0 - 15
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,    // 16 - 31
 _SK, _SK, _SK, _SK, _SK, _SK, 062, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, 063,    // 32 - 47
 052, 053, 054, 055, 056, 057, 058, 059, 060, 061, _SK, _SK, _SK, _EQ, _SK, _SK,    // 48 - 63
 _SK, 000, 001, 002, 003, 004, 005, 006, 007, 008, 009, 010, 011, 012, 013, 014,    // 64 - 79
 015, 016, 017, 018, 019, 020, 021, 022, 023, 024, 025, _SK, _SK, _SK, _SK, _SK,    // 80 - 95
 _SK, 026, 027, 028, 029, 030, 031, 032, 033, 034, 035, 036, 037, 038, 039, 040,    // 96 - 111
 041, 042, 043, 044, 045, 046, 047, 048, 049, 050, 051, _SK, _SK, _SK, _SK, _SK,    // 112 - 127
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK,
 _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK, _SK
 );

//------------------------------------------------------------------------------
// structures for handling referers

 var g_sysobjLock: TRTLCriticalSection;

 const REFERER_COUNT = 8;
 var g_arrstrReferers: array [0..REFERER_COUNT-1] of String; // assuming this array is initailized with all empty strings
 var g_intRefererIndex: Integer;

//------------------------------------------------------------------------------

 function CreateRandomScript: String; forward;

//------------------------------------------------------------------------------
// In:
// _pcData -> asciiz data to hide in the URL.
// Out:
// _pcReferer -> HX_Allocate'd asciiz value to be passed in a Referer request header field
// Returns:
// HX_Allocated'd asciiz value for the URI part of the URL, exactly script?query 
//
// this particular implementation returns
// random/Path/name.ext?ABCDEscrambledthenbase64encodeddata
// where A and B contain checksum, and C, D and E contain scrambling seeds

 function HX_EncodeURL(const _pcData: PChar; var _pcReferer: PChar): PChar;
 var intEncodedLength: Integer;
     intOffset, intCounter, intTgt: Integer;
     ptrScan: Pointer;
     intThisByte, intPrevByte: Byte;
     ptrScramble: PChar;
     intRun: Integer;
     ptrScrambledData: Pointer;
     intMulPrime, intAddPrime, intInitPrime: Integer;
     intMulSeed, intAddSeed, intInitSeed, intCSum1, intCSum2: Integer;
     intRefererLength: Integer;
     intDataLength: Integer;
     strRandomScript: String;
     intRandomScriptLength: Integer;

 begin

  // Note to implementers: depending on the DataOverride setting (or BytesInURL
  // more clearly) in the ini file, the size of the data passed may be in range
  // of few hundred bytes, so try to not expand the data much while encoding.

  // Note to implementers: you MUST add randomness to the URL you create.
  // This function could be called more than once for the same request, in an
  // attempt to fool proxy signature blocking. If you return same URL for the
  // same request data, HTTPort will retry in vain and fail after all attempts
  // are gone.

  // Note to implementers: the data being passed is a BASE64 encoded line
  // containing the following raw request:
  // Field1: Value1<CRLF>
  // Field2: Value2<CRLF>
  // ...
  // <CRLF>
  // i.e. similar to the HTTP request.

  // Note to implementers: This particular implementation scrambles textual data
  // with binary PRNG and then BASE64 encodes it again. Sort of stupid, redundant
  // and makes larger URL for no good reason, but it is the way it is. The way
  // seeds and checksums are prepended to the URL is weird too.

  // Note to implementers: returning valid Referer is encouraged, although
  // not required. Keep referer sane, ex. it'd be awkward for .jpg to be
  // a referer. This particular implementation uses simple round-robin approach
  // where URL number N has a URL referer of N - REFERER_COUNT + 1
  // Note that the referer, as well as the resulting URL must be relative.

  try

   Result := nil;
   if (_pcData = nil) or (not Assigned(@_pcReferer)) then Exit;
   _pcReferer := nil;

   intDataLength := StrLen(_pcData);

   intMulPrime := Random(64);                   // pick three random indice
   intAddPrime := Random(64);
   intInitPrime := Random(64);

   intMulSeed := URLPrimes[intMulPrime];        // fetch prime seeds from the array using random indice
   intAddSeed := URLPrimes[intAddPrime];
   intInitSeed := URLPrimes[intInitPrime];

   ptrScrambledData := HX_AllocateMemory(intDataLength);   // allocate memory for scramble buffer
   try

    Move(_pcData^, ptrScrambledData^, intDataLength);      // copy data

    ptrScramble := PChar(ptrScrambledData);     // start scrambling from here
    intCSum1 := 0;                              // initialize checksums (fletcher-style)
    intCSum2 := 0;

    for intRun := 0 to intDataLength - 1 do     // run through all the data, xoring it over and calculating the checksum
    begin
     intInitSeed := (intInitSeed * intMulSeed + intAddSeed) and $ff;                // seed is updated using simple random generator
     ptrScramble[intRun] := Char(Byte(ptrScramble[intRun]) xor Byte(intInitSeed));  // next byte is xored with seed
     Inc(intCSum1, Byte(ptrScramble[intRun]));                                      // resulting byte updates the checksums
     Inc(intCSum2, intCSum1);
    end;

    // note that scrambling does not change the length of the data
    // now proceed to URL64 encoding

    // _client_ does the encoding, so I see no need optimizing code in assembler
    // (as with _server_ decoding)

    strRandomScript := CreateRandomScript;
    intRandomScriptLength := Length(strRandomScript);

    intEncodedLength := intRandomScriptLength + 5 + ((intDataLength + 2) div 3) * 4;  // script, seeds/checksums, BASE64 padding

    Result := HX_AllocateMemory(intEncodedLength + 1);
    FillChar(Result^, intEncodedLength, '=');
    Move(PChar(strRandomScript)^, Result^, intRandomScriptLength);

    intOffset := 0;
    intPrevByte := 0;
    intCounter := 0;
    ptrScan := ptrScrambledData;
    intTgt := intRandomScriptLength + 5;

    while intOffset < intDataLength do
    begin

     intThisByte := Byte(ptrScan^);

     case intCounter of
     0: Result[intTgt] := URL64MAP[intThisByte shr 2];
     1: Result[intTgt] := URL64MAP[((intPrevByte and 3) shl 4) or (intThisByte shr 4)];
     2:
     begin
      Result[intTgt] := URL64MAP[((intPrevByte and 15) shl 2) or (intThisByte shr 6)];
      Inc(intTgt);
      Result[intTgt] := URL64MAP[intThisByte and 63];
     end;
     end; // case

     intCounter := (intCounter + 1) mod 3;
     intPrevByte := intThisByte;

     Inc(intOffset); Inc(Integer(ptrScan)); Inc(intTgt);

    end; // while

    case intCounter of
    1: Result[intTgt] := URL64MAP[(intPrevByte and 3) shl 4];
    2: Result[intTgt] := URL64MAP[(intPrevByte and 15) shl 2];
    end;

    // prepend checksums and seeds to encoded data, now, this is weird

    Result[intRandomScriptLength] := URL64MAP[intCSum1 mod 64];
    Result[intRandomScriptLength + 1] := URL64MAP[intCSum2 mod 64];
    Result[intRandomScriptLength + 2] := URL64MAP[intInitPrime];
    Result[intRandomScriptLength + 3] := URL64MAP[intMulPrime];
    Result[intRandomScriptLength + 4] := URL64MAP[intAddPrime];

    // null-terminate the resulting string

    Result[intEncodedLength] := #0;

    // get to referer stuff

    EnterCriticalSection(g_sysobjLock); // interlock access to global variables
    try

     // save generated URL as a future referer

     g_arrstrReferers[g_intRefererIndex] := Result;
     g_intRefererIndex := (g_intRefererIndex + 1) mod REFERER_COUNT;

     // pick current referer in a round-robin fashion

     intRefererLength := Length(g_arrstrReferers[g_intRefererIndex]);
     if intRefererLength > 0 then
     begin
      _pcReferer := HX_AllocateMemory(intRefererLength + 1);
      Move(PChar(g_arrstrReferers[g_intRefererIndex])^, _pcReferer^, intRefererLength + 1);
     end
     else
      _pcReferer := nil;

    finally
     LeaveCriticalSection(g_sysobjLock);
    end;

   finally
    HX_ReleaseMemory(ptrScrambledData);
   end;

  except
   Result := nil;
  end;

 end;

//------------------------------------------------------------------------------
// In:
// _pcURL -> asciiz URL just received by HTTHost, possibly including http:// and hostname
// Returns:
// HX_Allocated'd asciiz original decoded value

 function HX_DecodeURL(const _pcURL: PChar): PChar;
 var ptrScan: PChar;
     intPad, intLength: Integer;
     intBufLen: Integer;
     ptrBuf: Pointer;
     ptrResult: PChar;
     intMulPrime, intAddPrime, intInitPrime: Integer;
     intMulSeed, intAddSeed, intInitSeed: Integer;
     intResScan, intCSum1, intCSum2, intC1, intC2: Integer;
     intQueryLength: Integer;
     intDataLength: Integer;

 begin

  // Note to implementers: This is really the most critical function of all,
  // because it's a publicly available entry point to the _server_, and anyone in
  // the world can essentially call this function with the parameter of her choice.
  // HTTHost calls this function with the ad-hoc URL received from the client. The
  // only guarantee that you have here is that it's a null-terminated line of printable
  // ascii of some reasonable length.

  try

   Result := nil;
   if _pcURL = nil then Exit;

   ptrScan := StrPos(_pcURL, '?');                        // ignore useless random script name
   if ptrScan = nil then ptrScan := _pcURL else Inc(ptrScan);

   intQueryLength := StrLen(ptrScan);
   if intQueryLength < 5 then Exit;                       // at least 5 chars must present - seeds and checksums

   intCSum1 := URL64MAP_REV[Byte(ptrScan[0])];            // fetch and verify seeds and checksums
   intCSum2 := URL64MAP_REV[Byte(ptrScan[1])];
   intInitPrime := URL64MAP_REV[Byte(ptrScan[2])];
   intMulPrime := URL64MAP_REV[Byte(ptrScan[3])];
   intAddPrime := URL64MAP_REV[Byte(ptrScan[4])];

   if (intMulPrime > 63) or (intAddPrime > 63) or (intInitPrime > 63) or
      (intCSum1 > 63) or (intCSum2 > 63) then Exit;

   intMulSeed := URLPrimes[intMulPrime];                  // fetch prime seeds from array
   intAddSeed := URLPrimes[intAddPrime];
   intInitSeed := URLPrimes[intInitPrime];

   intBufLen := ((intQueryLength - 2) div 4) * 4;         // + 3 - 5, first 5 chars need not to be decoded
   Inc(ptrScan, 5);                                       // skip seeds and checksums

   intPad := 0;

   ptrBuf := HX_AllocateMemory(intBufLen);
   try

    // BASE64 decoding, first pass using reverse map lookup

    asm

     push edi
     push esi
     push ebx

     cld
     xor eax, eax
     mov ecx, intBufLen
     mov edi, ptrBuf
     mov edx, ecx
     shr ecx, 2
     or ecx, ecx
     jz @@no_stosd
     rep stosd
@@no_stosd:
     mov ecx, edx
     and ecx, 3
     jz @@no_stosb
     stosb
@@no_stosb:

     mov edx, ptrScan
     mov edi, ptrBuf
     lea esi, URL64MAP_REV
     mov ecx, intPad
@@translation_loop:
     movzx ebx, byte ptr [edx]
     mov bl, byte ptr [ebx + esi]
     cmp bl, _TR
     jae @@extra_char
@@conv_char:
     mov byte ptr [edi], bl
     inc edi
@@skip_char:
     inc edx
     jmp @@translation_loop
@@extra_char:
     cmp bl, _SK
     jz @@skip_char
     cmp bl, _EN
     jz @@translation_loop_over
     cmp bl, _EQ
     jnz @@skip_char
     inc ecx
     mov bl, 0
     jmp @@conv_char

@@translation_loop_over:
     mov intPad, ecx
     sub edi, ptrBuf
     mov intBufLen, edi

     pop ebx
     pop esi
     pop edi

    end; // asm

    if (intPad <= 2) and ((intBufLen and $3) = 0) and (intBufLen > 0) then // sanity check
    begin

     intLength := (intBufLen shr 2) * 3 - intPad;
     intDataLength := intLength + intPad;
     Result := HX_AllocateMemory(intDataLength + 1);

     ptrResult := PChar(Result);

     // BASE64 decoding, second pass using shifts and ors

     asm

      push edi
      push esi
      push ebx

      mov esi, ptrBuf
      mov edi, ptrResult
      mov ecx, intBufLen
      shr ecx, 2
      or ecx, ecx
      jz @@no_decode

@@decode_loop:
      mov eax, dword ptr [esi]

      mov ebx, eax
      mov edx, eax
      and ebx, 0000003fh
      and edx, 00003000h
      shl ebx, 2
      shr edx, 12
      or ebx, edx
      mov byte ptr [edi], bl
      inc edi

      mov ebx, eax
      mov edx, eax
      and ebx, 00000f00h
      and edx, 003c0000h
      shr ebx, 4
      shr edx, 18
      or ebx, edx
      mov byte ptr [edi], bl
      inc edi

      mov ebx, eax
      mov edx, eax
      and ebx, 00030000h
      and edx, 3f000000h
      shr ebx, 10
      shr edx, 24
      or ebx, edx
      mov byte ptr [edi], bl
      inc edi

      add esi, 4
      loop @@decode_loop

@@no_decode:

      pop ebx
      pop esi
      pop edi

     end;

     // fully BASE64 decoded data is in Result, descramble it using
     // known seeds and verify checksum 

     ptrResult := PChar(Result);                // descramble scan
     intC1 := 0; intC2 := 0;                    // initial checksums

     for intResScan := 0 to intLength - 1 do    // byte-by-byte descrambling
     begin
      Inc(intC1, Byte(ptrResult[intResScan]));  // update running checksums with scrambled byte
      Inc(intC2, intC1);
      intInitSeed := (intInitSeed * intMulSeed + intAddSeed) and $ff; // update seed using additive random number generator
      Byte(ptrResult[intResScan]) := Byte(ptrResult[intResScan]) xor Byte(intInitSeed); // descramble this byte
     end;

     if ((intC1 and 63) <> intCSum1) or ((intC2 and 63) <> intCSum2) then  // checksum mismatch
     begin
      HX_ReleaseMemory(Result);
      Result := nil;
     end
     else                                                                  // valid checksum, valid result
     begin
      Result[intLength] := #0;                                             // cut the result at the actual length
     end;

     // even if the URL passes the checks, we have no clue really if it comes from
     // a valid client or malicious attacker, moreover even if it's a reasonably looking
     // HTTHost internal request, it can be forged, and HTTHost must deal with it 

    end;

   finally
    HX_ReleaseMemory(ptrBuf);
   end;

  except
   Result := nil;
  end;

 end;

//------------------------------------------------------------------------------

 function CreateRandomString(const _intLength: Integer): String;
 var intInd: Integer;
 begin

  Result := StringOfChar(#0, _intLength);

  for intInd := 0 to _intLength - 1 do
   PChar(Result)[intInd] := Chr(Ord('a') + Random(26));

  if Random(2) = 0 then PChar(Result)[0] := UpCase(PChar(Result)[0]);

 end;

//------------------------------------------------------------------------------
// note that this function returns a string that ends with question mark

 const EXTENSIONS_COUNT = 3;
 const arrstrExtensions: array [0..EXTENSIONS_COUNT-1] of String = ('.php?', '.asp?', '.pl?');

 function CreateRandomScript: String;
 var intInd: Integer;
     intPathParts: Integer;
 begin

  Result := '';

  intPathParts := Random(3);

  for intInd := 1 to intPathParts do
   Result := Result + CreateRandomString(4 + Random(4)) + '/';

  Result := Result + CreateRandomString(4 + Random(4)) + arrstrExtensions[Random(EXTENSIONS_COUNT)];

 end;

//------------------------------------------------------------------------------

initialization
 Randomize;
 InitializeCriticalSection(g_sysobjLock);
finalization
 DeleteCriticalSection(g_sysobjLock);
end.
