unit httportU;

interface

 const HTTPORT_DLL = 'httport.dll';

//------------------------------------------------------------------------------

 // retcode constants

 const H_SUCCESS = 0;

 const H_ERROR_BASE = $10000;
 const H_ERROR_WINSOCK_FAILED = H_ERROR_BASE + 0;
 const H_ERROR_ALREADY_STARTED = H_ERROR_BASE + 1;
 const H_ERROR_NOT_STARTED = H_ERROR_BASE + 2;
 const H_ERROR_NO_FREE_SLOTS = H_ERROR_BASE + 3;
 const H_ERROR_UNKNOWN_OPTION = H_ERROR_BASE + 4;
 const H_ERROR_INVALID_PARAMETER = H_ERROR_BASE + 5;
 const H_ERROR_SOCKS_ALREADY_STARTED = H_ERROR_BASE + 6;
 const H_ERROR_SOCKS_NOT_STARTED = H_ERROR_BASE + 7;

 // h_setoption constants

 const H_OPTION_BASE = $20000;
 const H_OPTION_HOST_LIST = H_OPTION_BASE + 0;
 const H_OPTION_USER_INFO = H_OPTION_BASE + 1;
 const H_OPTION_PROXY = H_OPTION_BASE + 2;
 const H_OPTION_DATA_OVERRIDE = H_OPTION_BASE + 3;
 const H_OPTION_HEADER_FIELDS = H_OPTION_BASE + 4;

 // connection mode constants

 const CONN_USE_AUTO = 0;
 const CONN_USE_CONNECT = 1;
 const CONN_USE_REMOTE = 2;

//------------------------------------------------------------------------------

 // type definitions

 type TRetCode = Cardinal;

 type TMappingParameters = record                                              // this structure is passed to H_CreateMapping
                            intStrucSize: Integer;                             // [in] Must be set to sizeof(TMappingParameters)
                            wLocalPort: Word;                                  // [in] Local port to listen
                            pcRemoteHost: PChar;                               // [in] Pointer to ASCIIZ string with mapped host name or IP
                            wRemotePort: Word;                                 // [in] Mapped port
                            intAllowLocalOnly: Integer;                        // [in] If not zero, the listening port will be bound to 127.0.0.1
                            intMode: Integer;                                  // [in] Proxy bypass mode, one of CONN_USE_... constants
                            intMappingId: Integer;                             // [out] Returns the index for the newly created mapping
                           end;

 type PMappingParameters = ^TMappingParameters;

 type TOptions = array [0..$ffff] of PChar;                                    // structure for passing list of pointers to ASCIIZ strings to H_SetOption
      POptions = ^TOptions;

 type TMappingStats = record                                                   // this structure is passed to H_GetMappingStats
                       intStrucSize: Integer;                                  // [in] Must be set to sizeof(TMappingStats)
                       intConnections: Integer;                                // DEPRECATED
                       intKBytesTransferred: Integer;
                       intSockErrors: Integer;
                       intHostErrors: Integer;
                       intTimeoutErrors: Integer;
                       intProtocolErrors: Integer;
                       intMemoryErrors: Integer;
                       intConnectErrors: Integer;
                       intOtherErrors: Integer;
                       intFatalErrors: Integer;
                      end;

 type PMappingStats = ^TMappingStats;

 type TProxyStats = record                                                     // this structure is passed to H_GetHTTPStats
                     intStrucSize: Integer;                                    // [in] Must be set to sizeof(TProxyStats)
                     intProxyFaults: Integer;                                  // [out] DEPRECATED
                     intHostFaults: Integer;                                   // [out] DEPRECATED
                     intCurrentMode: Integer;                                  // [out] The 'worst' bypass mode currently in use for all mappings
                     intDataOverride: Integer;                                 // [out] If the proxy forbids passing custom request headers, bytes of data passed per URL
                    end;

 type PProxyStats = ^TProxyStats;

//------------------------------------------------------------------------------

 // function declarations

 type PCardinal = ^Cardinal;

 function H_Start: TRetCode; stdcall; external HTTPORT_DLL;
 function H_Stop: TRetCode; stdcall; external HTTPORT_DLL;

 function H_SetOption(const _intOption, _intCount: Integer; const _parrpcParams: POptions): TRetCode; stdcall; external HTTPORT_DLL;
 function H_CreateMapping(_pstrucParams: PMappingParameters): TRetCode; stdcall; external HTTPORT_DLL;
 function H_GetListeningPort(const _intIndex: Integer; _pdwResult: PCardinal): TRetCode; stdcall; external HTTPORT_DLL;
 function H_DestroyMapping(const _intIndex: Integer): TRetCode; stdcall; external HTTPORT_DLL;

 function H_GetMappingStats(const _intIndex: Integer; _pstrucStats: PMappingStats): TRetCode; stdcall; external HTTPORT_DLL; // DEPRECATED
 function H_GetProxyStats(const _pstrucStats: PProxyStats): TRetCode; stdcall; external HTTPORT_DLL;

 function H_CreateSocks(_pstrucParams: PMappingParameters): TRetCode; stdcall; external HTTPORT_DLL;
 function H_GetSocksListeningPort(_pdwResult: PCardinal): TRetCode; stdcall; external HTTPORT_DLL;
 function H_DestroySocks: TRetCode; stdcall; external HTTPORT_DLL;

 function H_ReadLog(var _pcLogLines: PChar): TRetCode; stdcall; external HTTPORT_DLL;
 function H_ReleaseMemory(_ptrRaw: Pointer): TRetCode; stdcall; external HTTPORT_DLL;

//------------------------------------------------------------------------------

implementation

end.

