unit sampleU;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, httportU, ExtCtrls, Registry, ScktComp;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    edtProxyUsername: TEdit;
    Label3: TLabel;
    edtProxyPassword: TEdit;
    Label4: TLabel;
    edtProxyHost: TEdit;
    edtProxyPort: TEdit;
    edtTargetHost: TEdit;
    Label5: TLabel;
    edtTargetPort: TEdit;
    Label6: TLabel;
    memoTerminal: TMemo;
    Label7: TLabel;
    edtData: TEdit;
    Label8: TLabel;
    btnConnect: TButton;
    Label9: TLabel;
    Label10: TLabel;
    edtHostHost: TEdit;
    edtHostPort: TEdit;
    btnStartLibrary: TButton;
    objSocket: TClientSocket;
    Label11: TLabel;
    edtUserAgent: TEdit;
    cbRunSocks: TCheckBox;
    edtSocksPort: TEdit;
    Timer1: TTimer;
    Label12: TLabel;
    btnResetSettings: TButton;
    procedure FormDestroy(Sender: TObject);
    procedure btnStartLibraryClick(Sender: TObject);
    procedure btnStopLibraryClick(Sender: TObject);
    procedure btnConnectClick(Sender: TObject);
    procedure btnDisconnectClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure edtDataKeyPress(Sender: TObject; var Key: Char);
    procedure objSocketConnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure objSocketDisconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure objSocketRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure edtSocksPortExit(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure btnResetSettingsClick(Sender: TObject);
  private
    boolProcessing: Boolean;
    boolStarted: Boolean;
    boolConnected: Boolean;
    boolConnecting: Boolean;
    strucMapping: TMappingParameters;
    intMappingIndex: Integer;
    intSavedTunnelingMode: Integer;
    intSavedBytesInURL: Integer;
    wSocksPort: Word;
    boolSaveSettings: Boolean;
  private
    procedure RefreshControls;
    procedure GrayOut(_objComponent: TComponent);
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

//------------------------------------------------------------------------------

procedure CheckRetCode(const _strContext: String; const _dwRetCode: TRetCode);
begin
 if _dwRetCode <> H_SUCCESS then
 begin
  raise Exception.CreateFmt('HTTPort library call %s returned %.8x', [_strContext, _dwRetCode]);
 end;
end;

//------------------------------------------------------------------------------

procedure TForm1.btnStartLibraryClick(Sender: TObject);
var arrpcOptions: array [0..3] of PChar;
    dwSocksPort: Cardinal;
    strBytesInURL: String;
begin

 if boolProcessing then Exit;
 boolProcessing := True;
 try

  // system parameters are shared by all tunneled connections and may not
  // be changed when there are live connections

  arrpcOptions[0] := PChar(edtProxyHost.Text); // set proxy address and port
  arrpcOptions[1] := PChar(edtProxyPort.Text);
  CheckRetCode('H_SetOption(H_OPTION_PROXY)', H_SetOption(H_OPTION_PROXY, 2, @arrpcOptions));

  arrpcOptions[0] := PChar(edtProxyUsername.Text); // set proxy authentication and user-agent
  arrpcOptions[1] := PChar(edtProxyPassword.Text);
  arrpcOptions[2] := PChar(edtUserAgent.Text);
  CheckRetCode('H_SetOption(H_OPTION_USER_INFO)', H_SetOption(H_OPTION_USER_INFO, 3, @arrpcOptions));

  arrpcOptions[0] := PChar(edtHostHost.Text + ':' + edtHostPort.Text); // set HTTHosts list, must not be empty
  arrpcOptions[1] := PChar(edtHostHost.Text + ':' + edtHostPort.Text); // there could be many different hosts,
  arrpcOptions[2] := PChar(edtHostHost.Text + ':' + edtHostPort.Text); // this is just an example
  arrpcOptions[3] := PChar(edtHostHost.Text + ':' + edtHostPort.Text);
  CheckRetCode('H_SetOption(H_OPTION_HOST_LIST)', H_SetOption(H_OPTION_HOST_LIST, 4, @arrpcOptions));

  strBytesInURL := IntToStr(intSavedBytesInURL);
  arrpcOptions[0] := PChar(strBytesInURL); // set bytes in URL loaded from registry
  CheckRetCode('H_SetOption(H_OPTION_DATA_OVERRIDE)', H_SetOption(H_OPTION_DATA_OVERRIDE, 1, @arrpcOptions));

  CheckRetCode('H_Start', H_Start);

  if cbRunSocks.Checked then
  begin

   strucMapping.intStrucSize := SizeOf(TMappingParameters);
   strucMapping.wLocalPort := StrToInt(edtSocksPort.Text);
   strucMapping.pcRemoteHost := '';
   strucMapping.wRemotePort := 0;
   strucMapping.intMode := intSavedTunnelingMode;
   strucMapping.intAllowLocalOnly := 1;

   CheckRetCode('H_CreateSocks', H_CreateSocks(@strucMapping));

   repeat
    CheckRetCode('H_GetSocksListeningPort', H_GetSocksListeningPort(@dwSocksPort));
    Sleep(0);
   until dwSocksPort > 0;

   wSocksPort := dwSocksPort;
   memoTerminal.Lines.Add(Format('SOCKS4 is enabled at 127.0.0.1:%d', [wSocksPort]));

  end;

  boolStarted := True;
  RefreshControls;

 finally
  boolProcessing := False;
 end;

end;

//------------------------------------------------------------------------------

procedure TForm1.btnStopLibraryClick(Sender: TObject);
begin

 if boolProcessing then Exit;
 boolProcessing := True;
 try

  CheckRetCode('H_Stop', H_Stop);
  boolStarted := False;
  boolConnected := False;
  boolConnecting := False;
  RefreshControls;

 finally
  boolProcessing := False;
 end;

end;

//------------------------------------------------------------------------------

procedure TForm1.btnConnectClick(Sender: TObject);
var dwBoundPort: Cardinal;
begin

 if boolProcessing then Exit;
 boolProcessing := True;
 try

  if intSavedTunnelingMode = CONN_USE_AUTO then
   memoTerminal.Lines.Add('Creating mapping using "AUTO" tunneling mode')
  else if intSavedTunnelingMode = CONN_USE_CONNECT then
   memoTerminal.Lines.Add('Creating mapping using "CONNECT" tunneling mode')
  else if intSavedTunnelingMode = CONN_USE_REMOTE then
   if intSavedBytesInURL = 0 then
    memoTerminal.Lines.Add('Creating mapping using "REMOTE HOST" tunneling mode, normal sending mode')
   else
    memoTerminal.Lines.Add(Format('Creating mapping using "REMOTE HOST" tunneling mode, sending %d bytes of data per URL', [intSavedBytesInURL]));

  strucMapping.intStrucSize := SizeOf(TMappingParameters);
  strucMapping.wLocalPort := 0;
  strucMapping.pcRemoteHost := PChar(edtTargetHost.Text);
  strucMapping.wRemotePort := StrToInt(edtTargetPort.Text);
  strucMapping.intMode := intSavedTunnelingMode;
  strucMapping.intAllowLocalOnly := 1;

  CheckRetCode('H_CreateMapping()', H_CreateMapping(@strucMapping));
  intMappingIndex := strucMapping.intMappingId;

  repeat
   CheckRetCode('H_GetListeningPort', H_GetListeningPort(intMappingIndex, @dwBoundPort));
   Sleep(0);
  until dwBoundPort > 0;

  memoTerminal.Lines.Add(Format('Tunnel entry is at 127.0.0.1:%d', [dwBoundPort]));
  memoTerminal.Lines.Add('Connecting');

  objSocket.Host := '127.0.0.1';
  objSocket.Port := dwBoundPort;
  objSocket.Active := True;

  boolConnecting := True;

  RefreshControls;

 finally
  boolProcessing := False;
 end;

end;

//------------------------------------------------------------------------------

procedure TForm1.btnDisconnectClick(Sender: TObject);
var strucProxyStats: TProxyStats;
    intOldSavedBytesInURL: Integer;
begin

 if boolProcessing then Exit;
 boolProcessing := True;
 try

  strucProxyStats.intStrucSize := SizeOf(TProxyStats);                      // fetch actual proxy settings before closing
  CheckRetCode('H_GetProxyStats', H_GetProxyStats(@strucProxyStats));       // the mapping
  intSavedTunnelingMode := strucProxyStats.intCurrentMode;
  intOldSavedBytesInURL := intSavedBytesInURL;
  intSavedBytesInURL := strucProxyStats.intDataOverride;

  if intSavedTunnelingMode = CONN_USE_AUTO then
   memoTerminal.Lines.Add('The actual tunneling mode was "AUTO"')
  else if intSavedTunnelingMode = CONN_USE_CONNECT then
   memoTerminal.Lines.Add('The actual tunneling mode was "CONNECT"')
  else if intSavedTunnelingMode = CONN_USE_REMOTE then
   if intSavedBytesInURL = 0 then
    memoTerminal.Lines.Add('The actual tunneling mode was "REMOTE HOST", normal sending mode')
   else
    memoTerminal.Lines.Add(Format('The actual tunneling mode was "REMOTE HOST", sending %d bytes of data per URL', [intSavedBytesInURL]));

  CheckRetCode('H_DestroyMapping', H_DestroyMapping(intMappingIndex));

  boolConnected := False;
  boolConnecting := False;

  RefreshControls;

  // This only needs to be checked once - if the proxy is paranoidal (cuts off
  // unknown header fields), connection attempt will fail on first send. BytesInURL
  // parameter will then be internally set to non-zero indicating that fact.
  // As some data has been probably received already, there is no way of restoring
  // connection context and so we silently retry. This only happens once, as next time
  // library will be using another approach for sending data, i.e. including data
  // in the URL. See readme.txt for more info.

  if (intOldSavedBytesInURL = 0) and (intSavedBytesInURL > 0) then
  begin
   memoTerminal.Lines.Add('Paranoidal proxy detected, you may need to retry connection attempt once');
  end;

 finally
  boolProcessing := False;
 end;
 
end;

//------------------------------------------------------------------------------

procedure TForm1.GrayOut(_objComponent: TComponent);
begin
 if _objComponent is TEdit then
 with _objComponent as TEdit do
 begin
  if Enabled then Color := clWindow else Color := clBtnFace;
 end
 else
 if _objComponent is TMemo then
 with _objComponent as TMemo do
 begin
  if Enabled then Color := clWindow else Color := clBtnFace;
 end;
end;

//------------------------------------------------------------------------------

procedure TForm1.RefreshControls;
begin

 edtProxyHost.Enabled := not boolStarted;
 edtProxyPort.Enabled := not boolStarted;
 edtProxyUsername.Enabled := not boolStarted;
 edtProxyPassword.Enabled := not boolStarted;
 edtHostHost.Enabled := not boolStarted;
 edtHostPort.Enabled := not boolStarted;
 edtSocksPort.Enabled := not boolStarted;
 cbRunSocks.Enabled := not boolStarted;
 edtUserAgent.Enabled := not boolStarted;
 btnResetSettings.Enabled := not boolStarted;

 edtTargetHost.Enabled := boolStarted and (not boolConnected) and (not boolConnecting);
 edtTargetPort.Enabled := boolStarted and (not boolConnected) and (not boolConnecting);
 memoTerminal.Enabled := boolStarted and boolConnected;
 edtData.Enabled := boolConnected;

 if boolStarted then
 begin
  btnStartLibrary.Caption := 'Stop library';
  btnStartLibrary.OnClick := btnStopLibraryClick;
  btnConnect.Enabled := True;
  if boolConnected or boolConnecting then
  begin
   btnConnect.Caption := 'Disconnect';
   btnConnect.OnClick := btnDisconnectClick;
  end
  else
  begin
   btnConnect.Caption := 'Connect';
   btnConnect.OnClick := btnConnectClick;
  end;
 end
 else
 begin
  btnStartLibrary.Caption := 'Start library';
  btnStartLibrary.OnClick := btnStartLibraryClick;
  btnConnect.Enabled := False;
  btnConnect.Caption := 'Connect';
 end;

 GrayOut(edtProxyHost);
 GrayOut(edtProxyPort);
 GrayOut(edtProxyUsername);
 GrayOut(edtProxyPassword);
 GrayOut(edtHostHost);
 GrayOut(edtHostPort);
 GrayOut(edtTargetHost);
 GrayOut(edtTargetPort);
 GrayOut(memoTerminal);
 GrayOut(edtData);
 GrayOut(edtSocksPort);
 GrayOut(edtUserAgent);

end;

//------------------------------------------------------------------------------

procedure TForm1.FormCreate(Sender: TObject);
var objReg: TRegistry;
begin

 try // load two configuration parameters from registry

  objReg := TRegistry.Create;
  if Assigned(objReg) then
  try

   objReg.RootKey := HKEY_CURRENT_USER;
   if objReg.OpenKeyReadOnly('\Software\Technology Networks\HTTPort\SDK Sample') then
   try
    intSavedTunnelingMode := objReg.ReadInteger('Tunneling mode');
    intSavedBytesInURL := objReg.ReadInteger('Data bytes in URL');
   finally
    objReg.CloseKey;
   end
   else
    raise Exception.Create('No registry settings');

  finally
   objReg.Free;
  end;

 except
  intSavedTunnelingMode := CONN_USE_AUTO; // defaults
  intSavedBytesInURL := 0;
 end;

 boolStarted := False;
 boolConnected := False;

 RefreshControls;

end;

//------------------------------------------------------------------------------

procedure TForm1.FormDestroy(Sender: TObject);
var objReg: TRegistry;
begin

 if not boolSaveSettings then Exit;

 objReg := TRegistry.Create; // save two configuration parameters in registry
 if Assigned(objReg) then
 try

  objReg.RootKey := HKEY_CURRENT_USER;
  if objReg.OpenKey('\Software\Technology Networks\HTTPort\SDK Sample', True) then
  try
   objReg.WriteInteger('Tunneling mode', intSavedTunnelingMode);
   objReg.WriteInteger('Data bytes in URL', intSavedBytesInURL);
  finally
   objReg.CloseKey;
  end
  else
  if objReg.CreateKey('\Software\Technology Networks\HTTPort\SDK Sample') then
  begin
   objReg.OpenKey('\Software\Technology Networks\HTTPort\SDK Sample', True);
   try
    objReg.WriteInteger('Tunneling mode', intSavedTunnelingMode);
    objReg.WriteInteger('Data bytes in URL', intSavedBytesInURL);
   finally
    objReg.CloseKey;
   end;
  end
  else
   raise Exception.Create('Failed to store settings to registry');

 finally
  objReg.Free;
 end;

end;

//------------------------------------------------------------------------------

procedure TForm1.edtDataKeyPress(Sender: TObject; var Key: Char);
begin

 if Key = #13 then
 begin
  objSocket.Socket.SendText(edtData.Text + #13#10);
  memoTerminal.Lines.Add('> ' + edtData.Text);
  edtData.Clear;
 end;

end;

//------------------------------------------------------------------------------

procedure TForm1.objSocketConnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
 memoTerminal.Lines.Add('Connected');
 boolConnecting := False;
 boolConnected := True;
 RefreshControls;
end;

//------------------------------------------------------------------------------

procedure TForm1.objSocketDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin

 if boolConnected then
 begin
  btnConnect.OnClick(btnConnect);
 end;

 memoTerminal.Lines.Add('Disconnected');
 
end;

//------------------------------------------------------------------------------

procedure TForm1.objSocketRead(Sender: TObject; Socket: TCustomWinSocket);
begin
 memoTerminal.Lines.Add('< ' + Trim(Socket.ReceiveText));
end;

//------------------------------------------------------------------------------

procedure TForm1.edtSocksPortExit(Sender: TObject);
begin
 try
  edtSocksPort.Text := IntToStr(StrToInt(Trim(edtSocksPort.Text)) and $ffff);
 except
  edtSocksPort.Text := '0';
 end;
end;

//------------------------------------------------------------------------------

procedure TForm1.Timer1Timer(Sender: TObject);
var pcLogMessage: PChar;
begin
 CheckRetCode('H_ReadLog', H_ReadLog(pcLogMessage));
 try
  if StrLen(pcLogMessage) > 0 then memoTerminal.Lines.Add(pcLogMessage);
 finally
  CheckRetCode('H_ReleaseMemory', H_ReleaseMemory(pcLogMessage));
 end;
end;

//------------------------------------------------------------------------------

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
 CanClose := True;
 boolSaveSettings := MessageDlg('Save current library settings to the registry so that they a restored at next startup ?', mtConfirmation, [mbYes, mbNo], 0) = mrYes;
 if boolSaveSettings then
  MessageDlg('Library settings will be saved to the registry under HKCU\Software\Technology Networks\HTTPort\SDK Sample',
             mtInformation, [mbOk], 0);
end;

//------------------------------------------------------------------------------

procedure TForm1.btnResetSettingsClick(Sender: TObject);
begin

 if MessageDlg('Library settings will be reset to their initial values (AUTO connect mode, default send mode). Ok to proceed ?',
               mtConfirmation, [mbYes, mbNo], 0) = mrYes then
 begin
  intSavedTunnelingMode := CONN_USE_AUTO;
  intSavedBytesInURL := 0;
 end;

end;

//------------------------------------------------------------------------------

end.

